# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/client.py
""" Soar client entrypoint.

Classes and functions for interacting with Soar in general. This module serves as the main entrypoint to the package.
Projects desiring to invoke a Soar instance should import this module and call :func:`soar.client.main`.

Examples:
    This will invoke a GUI instance, which will terminate when the main window is closed, and always return 0::

        from soar.client import main
        return_value = main()

    If invoking a headless instance, paths to brain and world files should be specified::

        from soar.client import main
        return_value = main(brain_path='path/to/brain.py', world_path='path/to/world.py', headless=True)

    In this case, the return value will be 1 if an exception occurred and 0 otherwise.

    Logging is handled via passing a path::

        from soar.client import main
        return_value = main(logfile='path/to/logfile')

    or, using a file-like object::

        from soar.client import main
        return_value = main(logfile=open('path/to/logfile', 'r+'))
    """
import os
import sys
import atexit
import json
import traceback as tb
from io import BytesIO
from queue import Queue
from threading import current_thread, main_thread

import soar.hooks
from soar import __version__
from soar.common import *
from soar.errors import *
from soar.controller import Controller
from soar.gui.soar_ui import SoarUI
from soar.gui.plot_window import PlotWindow
from soar.update import get_update_message
from soar.sim.geometry import *
from soar.sim.world import *

MAIN_THREAD = main_thread()  # Defines the main thread for later comparisons

brain = None
brain_path = None
__module_paths = []
__brain_modules = []
world = None
world_path = None
__world_modules = []
robot = None
controller = None
gui = None
logfile = None
step_duration = 0.1
realtime = True
options = None
plots = []
queue = Queue()


def empty_queue():  # Empty the queue
    global queue
    while not queue.empty():
        _ = queue.get()
        queue.task_done()


def tkinter_execute(func, after_idle=False):  # Run functions on Tk's main thread, synchronously
    def tk_wrap(*args, **kwargs):
        global gui, MAIN_THREAD
        if current_thread() != MAIN_THREAD:  # Only force Tk to call the function if not already on the main thread
            def return_exceptions(*args, **kwargs):  # Force a function to return any exception it raises
                try:
                    return_val = func(*args, **kwargs)
                except Exception as e:
                    return e
                else:
                    return return_val
            return_val = gui.synchronous_future(return_exceptions, *args, after_idle=after_idle, **kwargs)
            if isinstance(return_val, Exception):  # If the function returned an exception, raise it
                raise return_val
            return return_val
        else:  # If already on the main thread, just call the function
            return func(*args, **kwargs)
    return tk_wrap


def catch_exceptions(func):  # Catch exceptions that occur and notify the client, via futures
    def catch_wrap(*args, **kwargs):
        try:
            return_val = func(*args, **kwargs)
        except LoggingError as e:
            printerr('LoggingError: ' + str(e))
            future(LOGGING_ERROR)
            return EXCEPTION
        except SoarIOError as e:
            printerr('SoarIOError: ' + str(e))
            future(CONTROLLER_IO_ERROR)
            return EXCEPTION
        except GUIError as e:
            tb.print_exc()
            future(GUI_ERROR)
            return EXCEPTION
        except Exception as e:  # Any other exception (likely occurring in the brain) signifies controller failure
            tb.print_exc()
            future(CONTROLLER_FAILURE)
            return EXCEPTION
        else:  # Otherwise return the function's return value
            return return_val
    return catch_wrap


@catch_exceptions  # Logging that doesn't explode if something goes wrong
def log(obj, mode='a'):
    """ Log a serialized JSON object to a path or file-like object, adding a newline after.

    Args:
        obj: The object to be serialized.
        mode: The mode in which the file is to be opened.
    """
    global logfile
    # Recast any write errors as Soar errors
    if hasattr(logfile, 'write'):  # Try and treat logfile as a file-like object that has been opened
        try:
            logfile.write(json.dumps(obj) + '\n')
        except Exception as e:
            raise LoggingError(str(e))
    else:  # Otherwise treat it as a path, and open it
        try:
            with open(logfile, mode) as f:
                f.write(json.dumps(obj) + '\n')
        except (OSError, IOError) as e:
            raise LoggingError(str(e))


def future(name, *args, **kwargs):
    """ Adds a function with data for the client to execute in the future, by putting it on the client's internal queue.

    Note that if an exception or controller failure occurs, the future may never be executed as the client's queue will
    be purged.

    Certain futures accept an optional callback parameter.

    Args:
        name: The future to execute, defined in :mod:`soar.common`. and contained in `future_map`.
        *args: The variable length arguments to pass to the function.
        **kwargs: The keyword arguments to pass to the function.
    """
    global queue
    queue.put((name, args, kwargs))


# Loads a module from a path and returns its namespace, as well as any modules it loaded, as a list
def load_module(path, namespace=None):
    global __module_paths
    if namespace is None:
        # Input is replaced with a non-blocking dummy function.
        # The builtins are copied to the new namespace.
        namespace = {'input': lambda *args, **kwargs: print('Intercepted call to blocking input().',
                                                            file=sys.stderr), '__builtins__': __builtins__.copy()}
    path = os.path.abspath(path)  # Load from absolute paths

    namespace['__name__'] = os.path.splitext(os.path.basename(path))[0]  # Base name of the module
    module_dir = os.path.dirname(path)  # Module directory; this is added to sys.path for loading the modules
    before_load = sys.modules.copy()  # Make a copy of sys.modules to compare to later
    try:
        __module_paths.append(module_dir)
        sys.path.append(module_dir)
        code_object = compile(open(path, 'r').read(), path, 'exec')
        exec(code_object, namespace)  # Execute the module in an isolated namespace
    except Exception as e:  # Unload any loaded modules if, say, a syntax error occurred during load
        loaded = [modname for modname in sys.modules if modname not in before_load]  # List the modules that were loaded
        for modname in loaded:
            try:
                del sys.modules[modname]
            except KeyError:  # Assume that if there is a KeyError, the module is no longer loaded
                pass
        raise e  # Re-raise the exception
    finally:
        # Remove any module paths from sys.path
        for module_path in __module_paths:
            try:
                __module_paths.remove(module_path)
                sys.path.remove(module_path)
            except ValueError:  # If the path is not in sys.path, that's fine
                pass
        loaded = [modname for modname in sys.modules if modname not in before_load]  # List the modules that were loaded
    return namespace, loaded
    # Below is the old way of importing modules, as actual module objects, not namespaces
    # spec = importlib.util.spec_from_file_location(os.path.splitext(path)[0], path)
    # module = importlib.util.module_from_spec(spec)
    # spec.loader.exec_module(module)


def brain_setup():  # Setup a freshly loaded brain by wrapping the necessary methods
    global brain, gui, logfile, plots
    force_main_thread = False  # If this gets set to True, brain methods must run in the main thread

    # If the brain uses Tkinter and we're running in GUI mode, force brain methods to run in the main thread
    if gui and 'tkinter_hook' in brain:
        force_main_thread = True

    # Clear any plot objects that may have been listed by any previously loaded brain
    plots = []

    # Wrap brain methods, running them in Tk's thread if necessary
    for func in ['on_load', 'on_start', 'on_step', 'on_stop', 'on_shutdown']:
        if func in brain and callable(brain[func]):
            if force_main_thread:  # We hack a bit to run methods on the Tk thread, even when called from any thread
                brain[func] = tkinter_execute(brain[func])
        else:  # Make undefined functions do nothing
            brain[func] = lambda *args, **kwargs: None


def log_all_plots():  # Log any PlotWindow objects to the logfile. Use with caution.
    global plots
    for p in filter(lambda p: not p._destroyed, plots):  # Every plot that has not been destroyed
        image_bytes = BytesIO()
        p.save(image_bytes, format='png')
        log({'type': 'plot', 'data': image_bytes.getvalue().hex()})
        p._destroyed = True


def wrap_attrs(obj, names):  # Wrap object callables with the client exception decorator
    for name in names:
        if hasattr(obj, name) and callable(getattr(obj, name)):
            setattr(obj, name, catch_exceptions(getattr(obj, name)))


def nop(*args, callback=None, **kwargs):  # A NOP, typically used for callbacks
    if callback:
        callback()


def set_hooks(*args, **kwargs):  # Set the hooks that must be defined before a brain is ever loaded
    global gui, plots, MAIN_THREAD
    hooks = soar.hooks

    # Give the brain access to the mode soar is running in
    hooks.is_gui = lambda: gui is not None

    # First make sure all calls to PlotWindow use Toplevel() rather than Tk()
    PlotWindow._tk_started = True

    # Wrap PlotWindow, ensuring it plays nicely with Soar and Tk
    # This includes keeping track of open plots, attaching the window to the UI, and wrapping class methods
    plots = []
    class WrappedPlotWindow(PlotWindow):
        if gui:
            def __init__(self, title='Plotting Window', visible=True):
                # Wrap the class init so that it runs in the Tk mainloop
                if current_thread() != MAIN_THREAD:  # If not on the main thread, force it
                    tkinter_execute(PlotWindow.__init__)(self, title, visible=visible)
                else:  # If we're already on the main thread, do a normal init
                    PlotWindow.__init__(self, title, visible=visible)
                # Attach the window to the UI, and add it to the list of plots
                gui.attach_window(self)
                plots.append(self)

            def __getattribute__(self, name):
                if current_thread() != MAIN_THREAD:  # If not on the main thread
                    # Force accessing the attribute to run on the main thread
                    attr = tkinter_execute(PlotWindow.__getattribute__)(self, name)
                    # If the attribute is callable, wrap it so that it runs on the main thread
                    if callable(attr):
                        return tkinter_execute(attr)
                    return attr
                else:
                    return PlotWindow.__getattribute__(self, name)
        else:
            def __init__(self, title='Plotting Window', visible=True):
                # Ensure the window is not visible, and add the plot to the global list
                PlotWindow.__init__(self, title, visible=False)
                plots.append(self)
    sys.modules['soar.gui.plot_window'].PlotWindow = WrappedPlotWindow

    # Set the Tkinter hook to attach windows to the GUI
    if gui:
        hooks.tkinter_hook = gui.attach_window

    # Set the Tkinter wrap to execute widget callables on the Tk mainloop
    if gui:
        def tk_wrap(widget, linked=True):
            methods = [func for func in dir(widget) if callable(getattr(widget, func)) and not func.startswith('__')]
            for func_name in methods:
                func = getattr(widget, func_name)
                setattr(widget, func_name, tkinter_execute(func, after_idle=False))
            return hooks.tkinter_hook(widget, linked)
        hooks.tkinter_wrap = tk_wrap


def make_gui(*args, **kwargs):  # Creates the GUI and enters the main UI loop
    global gui
    gui = SoarUI(client_future=future, client_mainloop=mainloop)
    gui.mainloop()  # Enter the Tk event loop, which only ends when the GUI closes
    return True  # Exit client mainloop after GUI closes


def load_brain(path, *args, callback=None, silent=False, **kwargs):  # Loads a brain file, optionally calling a callback
    global brain, brain_path, robot, gui, __brain_modules, options
    brain_path = path
    for modname in __brain_modules:  # Try and delete previously loaded modules
        try:
            del sys.modules[modname]
        except KeyError:  # Assume if there is a KeyError that the module no longer exists
            pass
    brain, __brain_modules = load_module(brain_path)
    robot = brain['robot']
    if options is not None:
        robot.set_robot_options(**options)  # Set the robot options
    brain_setup()
    if not silent:
        print('LOAD BRAIN:', path)
    if callback:
        callback()


def load_world(path, *args, callback=None, silent=False, **kwargs):  # Loads a world file, optionally calling a callback
    global world, world_path, __world_modules
    world_path = path
    for modname in __world_modules:  # Try and delete previously loaded modules
        try:
            del sys.modules[modname]
        except KeyError:  # Assume if there is a KeyError that the module no longer exists
            pass
    world, __world_modules = load_module(world_path)
    world = world['world']  # Grab the actual world object, not the module
    if not silent:
        print('LOAD WORLD:', path)
    if callback:
        callback()


def gui_load_brain(path, *args, **kwargs):  # Simulates loading a brain through the GUI
    global gui
    gui.future(gui.loading)
    gui.brain_path = os.path.abspath(path)
    load_brain(os.path.abspath(path), callback=lambda: gui.future(gui.brain_ready))


def gui_load_world(path, *args, **kwargs):  # Simulates loading a world through the GUI
    global gui
    gui.future(gui.loading)
    gui.world_path = os.path.abspath(path)
    load_world(os.path.abspath(path), callback=lambda: gui.future(gui.world_ready))


def make_controller(*args, simulated=True, callback=None, **kwargs):  # Makes the controller and loads it
    global robot, brain, brain_path, world, world_path, controller, gui, logfile, step_duration, realtime
    if logfile:  # Write initial meta information to file
        log({'type': 'meta', 'simulated': simulated, 'version': __version__, 'brain': os.path.abspath(brain_path),
             'world': os.path.abspath(world_path)}, mode='w')
        controller_log = log
    else:
        controller_log = None
    controller = Controller(client_future=future, robot=robot, brain=brain, world=world, simulated=simulated, gui=gui,
                            step_duration=step_duration, realtime=realtime, log=controller_log)
    wrap_attrs(controller, ['load', 'run', 'step_thread', 'stop', 'shutdown'])
    if not simulated:  # If working with a real robot, register its shutdown to be called at exit
        atexit.register(shutdown_robot, robot)
    if controller.load() == EXCEPTION:
        return
    if callback:
        callback()


def make_world_canvas(world, *args, callback=None, **kwargs):  # Builds the canvas in the gui
    global gui
    if gui:
        gui.future(gui.make_world_canvas, world, callback=callback)


def draw(obj, *args, **kwargs):  # Draws one or more objects by placing them on the UI's draw queue
    global gui
    gui.future(gui.draw, obj)


def start_controller(*args, callback=None, **kwargs):
    global controller, gui
    if controller:
        if controller.run() == EXCEPTION:
            return
    if callback:
        callback()


def pause_controller(*args, callback=None, **kwargs):
    global controller
    if controller:
        controller.pause()
    if callback:
        callback()


def step_controller(n, *args, **kwargs):
    global controller
    if controller:
        controller.run(n=n)


def step_finished(*args, **kwargs):
    global gui, controller
    if gui:  # If there is a gui, notify it that the steps have completed.
        if controller and not controller.stopped:  # However only do so if the controller has not stopped.
            gui.future(gui.step_finished)
    else:  # Break out of the mainloop after the steps have finished
        return True


def stop_controller(*args, callback=None, **kwargs):
    global controller
    if controller:
        if controller.stop() == EXCEPTION:
            return
    if callback:
        callback()


def shutdown_controller(*args, **kwargs):
    global controller, logfile
    if controller:
        controller.shutdown()
        if logfile:  # Log any plots that were created
            log_all_plots()
        atexit.unregister(shutdown_robot)
    controller = None


def logging_error(*args, **kwargs):  # Called when a LoggingError occurs
    global gui
    if not gui:  # Quit out of the mainloop if not running in GUI mode
        raise SoarError('Unable to handle logging errors in headless mode')


def gui_error(*args, **kwargs):
    global gui
    if gui:
        gui.future(gui.gui_error)
    else:
        raise SoarError('GUIErrors should not occur in headless mode')


def controller_io_error(*args, **kwargs):  # Called after a SoarIOError occurs
    global brain, brain_path, world, world_path, robot, gui
    if gui:
        gui.future(gui.controller_io_error)
    else:  # Err out of the mainloop
        raise SoarError('Unable to handle SoarIOError in headless mode')


def controller_failure(*args, **kwargs):  # The controller has failed in an unanticipated way
    global controller
    empty_queue()  # Empty the client queue
    if controller:
        controller.failure()
    if gui:
        gui.future(gui.controller_failure)
        atexit.unregister(shutdown_robot)  # No need to shut down the robot at exit anymore
    else:
        raise SoarError('Unable to handle controller failures in headless mode')


def controller_complete(*args, **kwargs):  # Called when the simulation has completed
    global gui, controller, logfile, brain
    if controller:
        if controller.stop() == EXCEPTION:
            return
        if gui:
            gui.future(gui.stop_ready)
        if controller.shutdown() == EXCEPTION:
            return
        print('Simulation completed in', round(controller.elapsed, 3), '(simulated) seconds.')
        if logfile:
            if len(args) > 0 and args[0]:  # Log an object passed to sim_completed
                log(args[0])
            if 'PlotWindow' in brain:  # Log any plot window objects
                log_all_plots()
            log({'type': 'meta', 'completed': controller.elapsed})
        if not gui:  # Prepare to quit out of the mainloop if no gui
            return True


def shutdown_robot(robot):  # Called by the mainloop and at exit to ensure the robot is shutdown
    try:
        robot.on_shutdown()
    except Exception:  # Ignore errors
        pass


future_map = {MAKE_GUI: make_gui,
              LOAD_BRAIN: load_brain,
              LOAD_WORLD: load_world,
              MAKE_CONTROLLER: make_controller,
              DRAW: draw,
              START_CONTROLLER: start_controller,
              PAUSE_CONTROLLER: pause_controller,
              STEP_CONTROLLER: step_controller,
              STOP_CONTROLLER: stop_controller,
              SHUTDOWN_CONTROLLER: shutdown_controller,
              CONTROLLER_COMPLETE: controller_complete,
              CONTROLLER_IO_ERROR: controller_io_error,
              CONTROLLER_FAILURE: controller_failure,
              STEP_FINISHED: step_finished,
              MAKE_WORLD_CANVAS: make_world_canvas,
              GUI_LOAD_BRAIN: gui_load_brain,
              GUI_LOAD_WORLD: gui_load_world,
              NOP: nop,
              LOGGING_ERROR: logging_error,
              GUI_ERROR: gui_error,
              SET_HOOKS: set_hooks}
""" A mapping from future names to their actual functions. """


def mainloop():
    global gui
    while True:
        name, args, kwargs = queue.get()
        try:
            quit_loop = future_map[name](*args, **kwargs)
        except Exception as e:  # Catch any unhandled exceptions
            tb.print_exc()
            if gui:  # If there is a GUI, treat the exception as non-fatal and signal a controller failure
                controller_failure()
            else:  # Otherwise, return and signify an error
                return 1
        else:  # Successful client function execution
            if quit_loop:  # If a function returns true, end the loop and complete
                return 0
        finally:
            queue.task_done()


def main(brain_path=None, world_path=None, headless=False, logfile=None, step_duration=0.1, realtime=True,
         options=None):
    """ Main entrypoint, for use from the command line or other packages. Starts a Soar instance.

    Args:
        brain_path (optional): The path to the initial brain to load. Required if headless, otherwise not required.
        world_path (optional): The path to the initial world to load. Required if headless, otherwise not required.
        headless (bool): If `True`, run Soar in headless mode, immediately running a simulation.
        logfile (optional): The path to the log file, or a file-like object to log data to.
        step_duration (float): The duration of a controller step, in seconds. By default, 0.1 seconds.
        realtime (bool): If `True`, the controller will never sleep to make a step last the proper length. Instead,
            it will run as fast as possible.
        options (dict): The keyword arguments to pass to the robot whenever it is loaded.

    Returns:
        0 if Soar's execution successfully completed, 1 otherwise.
    """
    global robot
    globals()['logfile'] = logfile
    globals()['step_duration'] = step_duration
    globals()['realtime'] = realtime
    globals()['options'] = options
    if headless:
        future(SET_HOOKS)
        if brain_path and world_path:
            future(LOAD_BRAIN, brain_path)
            future(LOAD_WORLD, world_path)
        else:
            printerr('Both a brain and world file are required while running in headless mode.')
            return 1
        future(MAKE_CONTROLLER, simulated=True)
        future(START_CONTROLLER)
    else:
        future(MAKE_GUI)
        future(SET_HOOKS)
        if brain_path:
            future(GUI_LOAD_BRAIN, brain_path)
        if world_path:
            future(GUI_LOAD_WORLD, world_path)
    return_val = mainloop()
    shutdown_robot(robot)
    atexit.unregister(shutdown_robot)  # No need to shut down the robot at program exit anymore, since we're done
    return return_val



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/common.py
""" Soar common constants.

Contains named constants for sending messages to the client and determining the type of a controller.

Most use cases of soar will never need to use these constants or order client futures.

The values of the constants themselves are arbitrary.
"""
# Client futures
MAKE_GUI = 0  #: No arguments. Build the main Tkinter-based GUI and enter its event loop
LOAD_BRAIN = 1  #: Loads a brain, optionally calling the `callback` argument, and prints if `silent` is `False`.
LOAD_WORLD = 2  #: Loads a world, optionally calling the `callback` argument, and prints if `silent` is `False`.
MAKE_CONTROLLER = 3  #: Makes and loads the controller (type is based on `simulated`), and calls `callback`.
START_CONTROLLER = 4  #: Starts the controller, calling an optional callback.
PAUSE_CONTROLLER = 5  #: Pauses the controller, calling an optional callback.
STEP_CONTROLLER = 6  #: Steps the controller, where the first argument is the number of steps, infinite if `None`.
STOP_CONTROLLER = 7  #: Stops the controller and calls an optional callback.
SHUTDOWN_CONTROLLER = 8  #: Shuts down the controller.
CONTROLLER_COMPLETE = 9  #: Signal that the controller has completed and the simulation can end.
CONTROLLER_IO_ERROR = 11  #: Signals an IO error, which usually occurs when connecting with a real robot.
CONTROLLER_FAILURE = 12  #: Signals a controller failure.
LOGGING_ERROR = 13  #: Signals a logging error. These are typically ignored entirely.
STEP_FINISHED = 14  #: Signals that the step thread has finished.
GUI_LOAD_BRAIN = 15  #: Forces loading a brain as if it were done through the GUI.
GUI_LOAD_WORLD = 16  #: Forces loading a world as if it were done through the GUI.
SET_HOOKS = 17  #: Initializes the hooks
NOP = 18  #: Does nothing besides calling an optional callback

# Messages shared between the client and UI
DRAW = 19  #: Draw an object on the GUI's canvas.
MAKE_WORLD_CANVAS = 20  #: Tell the GUI to make the simulator canvas.
GUI_ERROR = 21  #: Signals an exception occurring somewhere in Tkinter callback

# Other constants
EXCEPTION = 22  #: Returned by wrapped functions to signal that an exception occurred during their execution.



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/gui/canvas.py
""" Soar canvas classes and methods.

Defines `SoarCanvas` and `SoarCanvasFrame` classes, as well as a method of creating them from an instance of
:class:`soar.sim.world.World`.
"""
import platform
from tkinter import Toplevel, BOTH, YES, Canvas, Frame

from soar import __version__

SYSTEM = platform.system()


def canvas_from_world(world, toplevel=Toplevel, close_cmd=None):
    """ Return a :class:`soar.gui.canvas.SoarCanvas` in a new window from a World. Optionally, call a different
    `toplevel()` method to create the window, and set its behavior on close.

    Additionally, if an object in the world defines `on_press`, `on_release`, and `on_motion` methods, bind those to
    the associated mouse events for that tag, making it draggable.

    Args:
        world: An instance of :class:`soar.sim.world.World` or a subclass.
        toplevel (optional): The function or method to call to create the window itself.
        close_cmd (optional): The function or method to call when the window is destroyed.

    Returns:
        The new `SoarCanvas`.
    """
    dim_x, dim_y = world.dimensions
    max_dim = max(dim_x, dim_y)
    width = int(dim_x / max_dim * 500)
    height = int(dim_y / max_dim * 500)
    options = {'width': width, 'height': height, 'pixels_per_meter': 500 / max_dim, 'bg': 'white'}
    t = toplevel()
    t.title('Soar ' + __version__ + ' Simulation')
    if close_cmd:
        t.protocol('WM_DELETE_WINDOW', close_cmd)
    t.aspect(width, height, width, height)
    f = SoarCanvasFrame(t)
    f.pack(fill=BOTH, expand=YES)
    c = SoarCanvas(f, **options)
    c.pack(fill=BOTH, expand=YES)
    world.canvas = c
    for obj in world:  # Bind mouse events to any object in the world which defines them
        for attr in ['on_press_left', 'on_motion_left', 'on_release_left', 'on_press_right', 'on_motion_right',
                     'on_release_right']:
            if hasattr(obj, attr):
                if attr == 'on_press_left':
                    c.tag_bind(obj.tags, '<ButtonPress-1>', obj.on_press_left)
                elif attr == 'on_motion_left':
                    c.tag_bind(obj.tags, '<B1-Motion>', obj.on_motion_left)
                elif attr == 'on_release_left':
                    c.tag_bind(obj.tags, '<ButtonRelease-1>', obj.on_release_left)
                elif attr == 'on_press_right':
                    if SYSTEM == 'Darwin':  # Right-click bindings on Mac are different
                        c.tag_bind(obj.tags, '<ButtonPress-2>', obj.on_press_right)
                    else:
                        c.tag_bind(obj.tags, '<ButtonPress-3>', obj.on_press_right)
                    # Also bind to Ctrl-click for Mac, in case there isn't a right click for some reason?
                    c.tag_bind(obj.tags, '<Control-ButtonPress-1>', obj.on_press_right)
                elif attr == 'on_motion_right':
                    if SYSTEM == 'Darwin':
                        c.tag_bind(obj.tags, '<B2-Motion>', obj.on_motion_right)
                    else:
                        c.tag_bind(obj.tags, '<B3-Motion>', obj.on_motion_right)
                        c.tag_bind(obj.tags, '<Control-B1-Motion>', obj.on_motion_right)
                elif attr == 'on_release_right':
                    if SYSTEM == 'Darwin':
                        c.tag_bind(obj.tags, '<ButtonRelease-2>', obj.on_release_right)
                    else:
                        c.tag_bind(obj.tags, '<ButtonRelease-3>', obj.on_release_right)
                    c.tag_bind(obj.tags, '<Control-ButtonRelease-1>', obj.on_release_right)
    return c


class SoarCanvas(Canvas):
    """ A metered, resizable Tkinter canvas. All drawing operations take metered arguments, where x values increase
    rightwards and y values increase upwards.

    Args:
        parent: The parent window or widget in which the canvas is placed.
        **options: Tk options.
    """
    def __init__(self, parent, **options):
        self.pixels_per_meter = options.pop('pixels_per_meter', 100)
        Canvas.__init__(self, parent, highlightthickness=0, **options)
        self.bind("<Configure>", self.on_resize)
        self.width = options['width']
        self.height = options['height']

    def on_resize(self, event):
        wscale = event.width / float(self.width)
        hscale = event.height / float(self.height)
        scale = min(wscale, hscale)
        self.width *= scale
        self.height *= scale
        self.config(width=self.width, height=self.height)
        self.pixels_per_meter *= scale
        self.scale('all', 0, 0, scale, scale)

    def create_arc(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_arc(self, *args, **kw)

    def create_bitmap(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_bitmap(self, *args, **kw)

    def create_image(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_image(self, *args, **kw)

    def create_line(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_line(self, *args, **kw)

    def create_oval(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_oval(self, *args, **kw)

    def create_polygon(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_polygon(self, *args, **kw)

    def create_rectangle(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_rectangle(self, *args, **kw)

    def create_text(self, *args, **kw):
        args = self.remap_coords(args)
        Canvas.create_text(self, *args, **kw)

    def remap_coords(self, coords):
        remapped = []
        for i in range(len(coords)):
            c = coords[i]*self.pixels_per_meter
            if i % 2 == 1:  # Every other coordinate is a y coordinate, and so must be remapped
                c = self.height-c
            remapped.append(c)
        return remapped


class SoarCanvasFrame(Frame):
    """ A resizable frame that holds a `SoarCanvas`. """
    def __init__(self, parent, **options):
        Frame.__init__(self, parent, **options)
        self.bind("<Configure>", self.on_resize)
        self.height = self.winfo_reqheight()
        self.width = self.winfo_reqwidth()

    def on_resize(self, event):
        self.width = event.width
        self.height = event.height
        for child in self.children.values():
            if isinstance(child, SoarCanvas):
                child.config(width=event.width, height=event.height)


#!/usr/bin/env python3
# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/__main__.py
""" Allows use of Soar from the command line by passing arguments to :func:`soar.client.main`
    ::

        usage: soar [-h] [--headless] [--nosleep] [--logfile LOGFILE]
                    [-s step duration] [-b brain] [-w world] [--options OPTIONS]

        optional arguments:
          -h, --help         show this help message and exit
          --headless         Run in headless mode
          --nosleep          Run quickly, without sleeping between steps
          --logfile LOGFILE  Log file to write to
          -s step duration   The duration of a controller step
          -b brain           Path to the brain file
          -w world           Path to the world file
          --options OPTIONS  Options to pass to the robot, as a JSON deserializable dictionary
"""
import sys
import json
from argparse import ArgumentParser

from soar import __version__, blerb
import soar.client as client


def main():
    parser = ArgumentParser(prog='soar', description=blerb)
    parser.add_argument('--headless', action='store_true', help='Run in headless mode')
    parser.add_argument('--nosleep', action='store_true', help='Run quickly, without sleeping between steps',
                        dest='nosleep')
    parser.add_argument('--logfile', type=str, help='Log file to write to', required=False)
    parser.add_argument('-s', metavar='step duration', type=float, help='The duration of a controller step',
                        required=False, default=0.1, dest='step_duration')
    parser.add_argument('-b', metavar='brain', type=str, help='Path to the brain file', required=False,
                        dest='brain_path')
    parser.add_argument('-w', metavar='world', type=str, help='Path to the world file', required=False,
                        dest='world_path')
    parser.add_argument('--options', type=str, help='Options to pass to the robot, as a JSON deserializable dictionary',
                        required=False)
    args = parser.parse_args()
    if args.options is not None:
        args.options = json.loads(args.options)
        assert(type(args.options) is dict)  # Make sure the options are valid
    return_val = client.main(brain_path=args.brain_path, world_path=args.world_path, headless=args.headless,
                             logfile=args.logfile, step_duration=args.step_duration, realtime=not args.nosleep,
                             options=args.options)
    sys.exit(return_val)

if __name__ == '__main__':
    main()



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/gui/output.py
""" Soar output classes.

Tk widgets allowing the capture and display of text output in the GUI.
"""
import webbrowser
from io import StringIO

from tkinter import *


class SoarIO(StringIO):
    def __init__(self, write_func):
        StringIO.__init__(self)
        self.write_func = write_func

    def write(self, s):
        self.write_func(s)


_link_id = 0


def _new_link_id():
    global _link_id
    return 'link' + str(_link_id)
    _link_id += 1


class OutputFrame(Frame):
    """ A read-only Tk output frame that can display normal and error output.

    Args:
        master: The parent widget or window.
        initial_text (str, optional): The text to have initially in the output frame.
    """
    def __init__(self, master, initial_text=''):
        Frame.__init__(self, master, bg='white')
        self.scroll = Scrollbar(self)
        self.scroll.pack(side=RIGHT, fill=Y)
        self.text_field = Text(self, fg='black', bg='white', wrap=WORD, yscrollcommand=self.scroll.set)
        self.text_field.config(height=16, padx=5, pady=5, state=DISABLED)
        self.text_field.tag_config('output', foreground='black')
        self.text_field.tag_config('error', foreground='red')
        self.text_field.bind("<1>", lambda event: self.text_field.focus_set())
        self.text_field.pack(expand=True, fill='both')
        self.scroll.config(command=self.text_field.yview)
        if initial_text != '':
            self.output(initial_text)

    def insert(self, text, *tags):
        """ Insert text at the end of the text field.

        Args:
            text (str): The text to insert.
            *tags: Variable length `str` list of tags to attach to the text. The `'output'` tag signifies normal output,
                   and the `'error'` tag signifies that the text will be red.
        """
        self.text_field.config(state=NORMAL)
        self.text_field.insert(END, text, *tags)
        self.text_field.see("%s-2c" % END)  # Ensure the inserted text is visible
        self.text_field.config(state=DISABLED)

    def output(self, text):
        """ Insert normal output text at the end of the text field.

        Args:
            text (str): The text to insert.
        """
        self.insert(text, 'output')

    def error(self, text):
        """ Insert error output text at the end of the text field.

        Args:
             text (str): The text to insert.
        """
        self.insert(text, 'error')

    def link(self, text):
        """ Insert a clickable link at the end of the text field.

        Args:
            text (str): The link text to insert.
        """
        tag = _new_link_id()
        self.text_field.tag_config(tag, foreground='blue', underline=1)
        self.text_field.tag_bind(tag, '<Button-1>', lambda event: webbrowser.open_new(text))
        self.text_field.tag_bind(tag, '<Enter>', lambda event: self.text_field.config(cursor='hand1'))
        self.text_field.tag_bind(tag, '<Leave>', lambda event: self.text_field.config(cursor=''))
        self.insert(text, tag)

    def clear(self):
        """ Clear the entire text field. """
        self.text_field.config(state=NORMAL)
        self.text_field.delete(1.0, END)
        self.text_field.config(state=DISABLED)



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/robot/pioneer.py
""" A PioneerRobot class, for representing a real or simulated Pioneer 3 robot.

See the `MobileRobots documentation`_ for more information.

.. _MobileRobots documentation: http://www.mobilerobots.com/ResearchRobots/PioneerP3DX.aspx
"""
from math import pi, sqrt, atan2
from uuid import getnode

from soar.errors import SoarIOError
from soar.sim.geometry import Line, Pose, clip
from soar.sim.world import Polygon, Ray
from soar.robot.base import BaseRobot
from soar.robot.names import name_from_sernum
from soar.robot.arcos import *


class PioneerRobot(BaseRobot):
    """ An abstract, universal Pioneer 3 robot. Instances of this class can be fully simulated, or used to communicate
    with an actual Pioneer3 robot over a serial port.

    Attributes:
        type (str): Always `'Pioneer3'`; used to identify this robot type.
        simulated (bool): If `True`, the robot is being simulated. Otherwise it should be assumed to be real.
        pose: An instance of :class:`soar.sim.geometry.Pose` representing the robot's `(x, y, theta)` position.
             In simulation, this is the actual position; on a real robot this is based on information from the encoders.
        world: An instance of :class:`soar.sim.world.World` or a subclass, or `None`, if the robot is real.
        FV_CAP (float): The maximum translational velocity at which the robot can move, in meters/second.
        RV_CAP (float): The maximum rotational velocity at which the robot can move, in radians/second.
        SONAR_MAX (float): The maximum distance that the sonars can sense, in meters.
        arcos: An instance of :class:`soar.robot.arcos.ARCOSClient` if the robot is real and has been loaded, otherwise
               `None`.

    Args:
        **options: See `set_robot_options`.
    """
    def __init__(self, **options):
        self.FV_CAP = 1.5  # meters/second
        self.RV_CAP = 2 * pi  # radians/second
        BaseRobot.__init__(self, polygon=Polygon([(-0.034, -0.219), (0.057, -0.219), (0.139, -0.170), (0.196, -0.098),
                                                  (0.216, -0.009), (0.196, 0.079), (0.139, 0.150), (0.057, 0.190),
                                                  (-0.034, 0.190), (-0.051, 0.190), (-0.051, 0.150), (-0.153, 0.150),
                                                  (-0.223, 0.074), (-0.223, -0.103), (-0.153, -0.179), (-0.051, -0.179),
                                                  (-0.051, -0.219)], None, fill='black', dummy=True))
        self.type = 'Pioneer3'
        # Sonar poses relative to the robot's center
        self.sonar_poses = [Pose(0.08, 0.134, pi/2), Pose(0.122, 0.118, 5*pi/18),
                            Pose(0.156, 0.077, pi/6), Pose(0.174, 0.0266, pi/18),
                            Pose(0.174, -0.0266, -pi/18), Pose(0.156, -0.077, -pi/6),
                            Pose(0.122, -0.118, -5*pi/18), Pose(0.08, -0.134, -pi/2)]
        self.SONAR_MAX = 1.5  # meters
        self.arcos = None  # The ARCOS client, if connected
        self._fv = 0.0  # Internal forward velocity storage
        self._rv = 0.0  # Internal rotational velocity storage
        self._collided = False  # Private flag to check if the robot has collided
        self._sonars = None  # Super secret calculated sonars (shh)
        self._move_data = {'x': 0, 'y': 0, 'item': None}  # Drag data for the canvas
        self._turn_data = {'x': 0, 'y': 0, 'item': None}  # Drag data for the canvas
        self._serial_ports = None  # Serial ports to use when connecting with ARCOS; any if None
        self._last_x, self._last_y = 0, 0  # For dealing with encoder rollover
        self._x_sum, self._y_sum = 0, 0  # For dealing with encoder rollover
        self._ignore_brain_lag = False  # For making the step duration fixed
        self.set_robot_options(**options)  # Sets any options passed in on construction

    def to_dict(self):
        """ Return a dictionary representation of the robot, usable for serialization.

        This contains the robot type, position, sonar data, and forward and rotational velocities.
        """
        d = BaseRobot.to_dict(self)
        d.update({'sonars': self.sonars, 'collided': self._collided})
        return d

    def set_robot_options(self, **options):
        """ Set Pioneer3 specific options. Any unsupported keywords are ignored.

        Args:
            **options: Arbitrary robot options.
            serial_ports (list, optional): Sets the serial ports to try connecting to with the ARCOS client.
            ignore_brain_lag (bool): If `True`, a step will always be assumed to be 0.1 seconds long. Otherwise,
                whatever duration the controller tells the robot to step is how long a step lasts.
        """
        if 'serial_ports' in options:
            self._serial_ports = options['serial_ports']
        if 'ignore_brain_lag' in options:
            self._ignore_brain_lag = options['ignore_brain_lag']

    @property
    def fv(self):
        """ `float` The robot's current translational velocity, in meters/second.

        Positive values indicate movement towards the front of the robot, and negative values indicate movement
        towards the back.

        Setting the robot's forward velocity is always subject to :attr:`soar.robot.pioneer.PioneerRobot.FV_CAP`.
        On a real robot, this is further limited by the hardware translational velocity cap.
        """
        return self._fv

    @fv.setter
    def fv(self, value):
        self._fv = clip(value, -self.FV_CAP, self.FV_CAP)
        if not self.simulated:  # For real robots, must send an ARCOS command
            try:
                self.arcos.send_command(VEL, int(self._fv*1000))  # Convert m/s to mm/s
            except Timeout as e:  # Recast arcos errors as Soar errors
                raise SoarIOError(str(e)) from e

    @property
    def rv(self):
        """ `float` The robot's current rotational velocity, in radians/second.

        Positive values indicate counterclockwise rotation (when viewed from above) and negative values indicate
        clockwise rotation.

        Setting the robot's rotational velocity is always subject to :attr:`soar.robot.pioneer.PioneerRobot.RV_CAP`.
        On a real robot, this is further limited by the hardware rotational velocity cap.
        """
        return self._rv

    @rv.setter
    def rv(self, value):
        self._rv = clip(value, -self.RV_CAP, self.RV_CAP)
        if not self.simulated:
            try:
                self.arcos.send_command(RVEL, int(self._rv * 180 / pi))  # Convert radians/sec to degrees/sec
            except Timeout as e:  # Recast arcos errors as Soar errors
                raise SoarIOError(str(e)) from e

    @property
    def sonars(self):
        """ (`list` of `float`) The latest sonar readings as an array.

        The array contains the latest distance sensed by each sonar, in order, clockwise from the robot's far left to
        its far right. Readings are given in meters and are accurate to the millimeter. If no distance was sensed by a
        sonar, its entry in the array will be `None`.
        """
        if self.simulated:  # If simulating, grab the data from the latest calculated sonars
            if not self._sonars:  # If somehow this has been called before sonars are calculated, calculate them
                self.calc_sonars()
            return [round(s, 3) if s < self.SONAR_MAX else None for s in self._sonars]
        else:  # Otherwise grab the sonar data from the ARCOS Client
            return [s/1000.0 if s != 5000 else None for s in self.arcos.sonars[:8]]  # Convert mm to meters

    @property
    def analogs(self):
        """ Get the robot's 4 analog inputs, so long as it is real and not simulated, as a 4 tuple. """
        if self.simulated:
            raise SoarIOError('Cannot access the inputs of a simulated robot (yet)')  # TODO: CMax integration
        else:
            # Assume that the io information is relatively current
            return tuple([round(a*10.0/1023.0, 3) for a in self.arcos.io['ANALOGS'][4:]])

    def set_analog_voltage(self, v):
        """ Sets the robot's analog output voltage.

        Args:
            v (float): The output voltage to set. This is limited to the range of 0-10V.
        """
        if self.simulated:
            raise SoarIOError('Cannot set the analog output of a simulated robot (yet)')  # TODO: CMax integration
        else:  # Sent the analog-digital output required to read it later
            try:
                self.arcos.send_command(DIGOUT, int(clip(v, 0, 10)*25.5) | 0xff00)
            except Timeout as e:  # Recast arcos errors as Soar errors
                raise SoarIOError(str(e)) from e

    def get_distance_right(self):
        """ Get the perpendicular distance to the right of the robot.

        Returns:
            float: The perpendicular distance to the right of the robot, assuming there is a linear surface.
        """
        return self.get_distance_right_and_angle()[0]

    def get_distance_right_and_angle(self):
        """ Get the perpendicular distance and angle to a surface on the right.

        Returns:
            `(d, a)` where `d` is the perpendicular distance to a surface on the right, assuming it is linear, and `a`
            is the angle to that surface.
        """
        # Build a list of the points the sonars hit, or None if the distance was greater than the sonar max
        endpoints = [Ray(origin, d, dummy=True).p2 if d else None for (origin, d) in zip(self.sonar_poses, self.sonars)]
        return self.dist_and_angle(endpoints[6], endpoints[7])

    def get_distance_left(self):
        """ Get the perpendicular distance to the left of the robot.

        Returns:
            float: The perpendicular distance to the left of the robot, assuming there is a linear surface.
        """
        return self.get_distance_left_and_angle()[0]

    def get_distance_left_and_angle(self):
        """ Get the perpendicular distance and angle to a surface on the left.

        Returns:
            `(d, a)` where `d` is the perpendicular distance to a surface on the left, assuming it is linear, and `a`
            is the angle to that surface.
        """
        # Build a list of the points the sonars hit, or None if the distance was greater than the sonar max
        endpoints = [Ray(origin, d, dummy=True).p2 if d else None for (origin, d) in zip(self.sonar_poses, self.sonars)]
        return self.dist_and_angle(endpoints[0], endpoints[1])

    def dist_and_angle(self, h0, h1):  # Used to calculate perpendicular distance from surfaces
        if h0 and h1:
            l = Line(h0, h1, normalize=True)  # It is *essential* that the lines be normalized
            l_x, l_y, l_d = l.a, l.b, l.c
            return abs(l_d), pi / 2 - atan2(l_y, l_x)
        elif h0:
            (hx, hy) = h0
            return sqrt(hx * hx + hy * hy), None
        elif h1:
            (hx, hy) = h1
            return sqrt(hx * hx + hy * hy), None
        else:
            return self.SONAR_MAX, None

    def calc_sonars(self):  # Calculate the actual sonar ranges. Called once per simulated timestep
        self._sonars = [0]*len(self.sonar_poses)
        for i in range(len(self.sonar_poses)):
            # We take each sonar and build a ray longer than the world's max diagonal
            origin = self.sonar_poses[i]
            # Translate and turn by the robot's pose, then rotate about its center
            origin = origin.transform(self.pose).rotate(self.pose.point(), self.pose[2])
            sonar_ray = Ray(origin, 1.5*max(self.world.dimensions), dummy=True)

            # Find all collisions with objects that aren't the robot itself
            # Sonars only accurate to the millimeter, so let epsilon be 0.001 meters
            collisions = self.world.find_all_collisions(sonar_ray, condition=lambda obj: obj is not self, eps=1e-3)
            if collisions:  # Should always be True since the world has boundaries
                # Sort the collisions by distance to origin
                distances = [origin.distance(p) for _, p in collisions]
                distances.sort()
                self._sonars[i] = distances[0]  # Sonar reading is the distance to the nearest collision

    def draw(self, canvas):  # Draw the robot
        BaseRobot.draw(self, canvas)
        self.draw_sonars(canvas)

    def draw_sonars(self, canvas):  # Draw just the sonars
        canvas.delete(self.tags + 'sonars')  # Deleting a nonexistent tag is safe, so always delete the sonar lines
        if not self._sonars:
            self.calc_sonars()
        for dist, pose in zip(self._sonars, self.sonar_poses):
            origin = pose.transform(self.pose).rotate(self.pose.point(), self.pose[2])
            fill = 'firebrick2' if dist > self.SONAR_MAX else 'gray'
            sonar_ray = Ray(origin, dist, tags=self.tags+'sonars', fill=fill, width=1)
            sonar_ray.draw(canvas)

    def delete(self, canvas):  # TODO: Deprecate this in 2.0
        canvas.delete(self.tags, self.tags + 'sonars')  # Delete both the robot and sonar tags

    def check_if_collided(self):  # Check if the robot has collided, and set its collision flag accordingly
        if self.simulated:
            for obj in self.world:  # Check for collisions
                if obj is not self:
                    if self.collision(obj):  # If any collision occurs
                        self._collided = True
                        self.polygon.options['fill'] = 'red'
                        return True
            self._collided = False
            self.polygon.options['fill'] = 'black'
            return False
        else:
            return False

    def on_load(self):
        if self.simulated:
            self.arcos = None
            print('Connected to Pioneer p3dx-sh MIT_0042 \'Denny\' (12.0V) [Simulated]')  # Hi Denny Freeman!
        else:
            try:
                self.arcos = ARCOSClient()
                self.arcos.connect(forced_ports=self._serial_ports)
                self.arcos.send_command(ENABLE, 0)  # We disable the motors so that the robot is easily movable
                # Continuously request IOpacs, and make sure we receive at least one
                self.arcos.send_command(IOREQUEST, 2)
                self.arcos.wait_or_timeout(self.arcos.io_event, 1.0, 'Could not access robot IO information')
                # Request a CONFIGpac and make sure we receive one
                self.arcos.send_command(CONFIG)
                self.arcos.wait_or_timeout(self.arcos.config_event, 1.0, 'Could not access robot configuration')
                config = self.arcos.config
                serial_num = str(getnode())
                battery_volts = self.arcos.standard['BATTERY'] / 10.0
                # Print the standard connection message and warn if the battery is low
                print('Connected to ' + ' '.join([config[field] for field in ['ROBOT_TYPE', 'SUBTYPE', 'NAME']])
                      + ' \'' + name_from_sernum(serial_num) + '\' (' + str(battery_volts) + ')')
                if battery_volts < self.arcos.config['LOWBATTERY']/10.0:
                    printerr('WARNING: The robot\'s battery is low. Consider recharging or finding a new one.')
            except ARCOSError as e:  # If anything goes wrong, raise a SoarIOError
                raise SoarIOError(str(e)) from e

    def on_start(self):
        if not self.simulated:
            for _ in range(5):  # Try enabling motors a few times
                if self.arcos.standard['FLAGS'] & 0x1 != 0x1:  # If the motors have not been enabled
                    try:
                        self.arcos.send_command(ENABLE, 1)  # Re-enable them
                    except Timeout as e:  # Recast arcos errors as Soar errors
                        raise SoarIOError(str(e)) from e
                    sleep(1.0)
                else:  # If they have been, we're all set
                    break
            if self.arcos.standard['FLAGS'] & 0x1 != 0x1:  # If they still aren't enabled, raise an error
                raise SoarIOError('Unable to enable the robot\'s motors')
            try:
                self.arcos.send_command(SETO)
            except Timeout as e:  # Recast arcos errors as Soar errors
                raise SoarIOError(str(e)) from e

    def on_step(self, duration):
        if self._ignore_brain_lag:
            duration = 0.1  # If ignoring brain lag, fix the step duration
        if self.simulated:  # Move, check for collisions, and update the internal sonars
            if not self._collided:  # The robot only moves if it hasn't collided
                BaseRobot.on_step(self, duration)  # Do BaseRobot's simulated move and collision preemption
            self.check_if_collided()
            self.calc_sonars()
        else:  # Update the position information from the encoders
            # Assume that io.standard is relatively current, which is likely
            # Encoders can roll over, so keep track of the last x, y encoder reports and update accordingly
            x, y, t = self.arcos.standard['XPOS'], self.arcos.standard['YPOS'], self.arcos.standard['THPOS']
            d_x, d_y = x-self._last_x, y-self._last_y
            if d_x > 60000:  # A positive rollover occurred
                d_x -= 65536
            elif d_x < -60000:  # A negative rollover occurred
                d_x += 65536
            if d_y > 60000:  # A positive rollover occurred
                d_y -= 65536
            elif d_y < -60000:  # A negative rollover occurred
                d_y += 65536
            self._last_x, self._last_y = x, y
            self._x_sum += d_x
            self._y_sum += d_y
            self.pose = Pose(self._x_sum/1000.0, self._y_sum/1000.0, t*0.001534)

    def on_stop(self):
        if not self.simulated:
            try:
                self.arcos.send_command(STOP)  # Stop the robot from moving
                self.arcos.send_command(ENABLE, 0)  # Disable the motors after we've stopped
            except Timeout as e:  # Recast arcos errors as Soar errors
                raise SoarIOError(str(e)) from e

    def on_shutdown(self):
        if self.arcos:
            self.arcos.disconnect()

    # Mouse event bindings follow
    # Left-click-drag moves the robot, while right-click-drag rotates it

    def on_press_left(self, event):
        self._move_data['x'] = event.x
        self._move_data['y'] = event.y

    def on_release_left(self, event):
        self._move_data['x'] = 0
        self._move_data['y'] = 0

    def on_motion_left(self, event):
        delta_x = event.x - self._move_data["x"]
        delta_y = event.y - self._move_data["y"]
        self._move_data["x"] = event.x
        self._move_data["y"] = event.y
        real_d_x = delta_x / self.world.canvas.pixels_per_meter
        real_d_y = -delta_y / self.world.canvas.pixels_per_meter
        # Update the robot's real position, check for a collision, and redraw/recalcuate the robot and sonars
        self.pose = self.pose.transform((real_d_x, real_d_y, 0))
        self.polygon.recenter(self.pose)
        if self.check_if_collided():
            self.world.canvas.itemconfigure(self.tags, fill='red')
        else:
            self.world.canvas.itemconfigure(self.tags, fill='black')
        self.calc_sonars()
        self.draw(self.world.canvas)

    def on_press_right(self, event):
        self._turn_data['x'] = event.x
        self._turn_data['y'] = event.y

    def on_release_right(self, event):
        self._turn_data['x'] = 0
        self._turn_data['y'] = 0

    def on_motion_right(self, event):
        delta_x = event.x - self._turn_data["x"]
        delta_y = event.y - self._turn_data["y"]
        self._turn_data["x"] = event.x
        self._turn_data["y"] = event.y
        # Moving left or up rotates counterclockwise, right or down rotates clockwise
        theta = -delta_x*2*pi/150-delta_y*2*pi/150
        # Change the robot's position, rotate and redraw the polygon, and recalculate and redraw the sonars
        self.pose = self.pose.transform((0, 0, theta))
        self.polygon.rotate(self.polygon.center, theta)
        self.calc_sonars()
        self.draw(self.world.canvas)




# This file is intended as an example. Use it as a template for other Soar brains
from soar.robot.base import BaseRobot
from soar.gui.plot_window import PlotWindow
from soar.hooks import sim_completed
from soar.sim.world import Polygon

robot = BaseRobot(polygon=Polygon([(-0.5, 0.5), (0.5, 0.5), (0.5, -0.5), (-0.5, -0.5)], tags='base'))


#  This function is called when the brain is loaded
def on_load():
    pass


#  This function is called when the start button is pushed
def on_start():
    pass


#  This function is called every step_duration seconds. By default, it is called 10 times/second
def on_step(step_duration):
    pass


# This function is called when the stop button is pushed
def on_stop():
    pass


# This function is called when the robot's controller is shut down
def on_shutdown():
    pass



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/sim/geometry.py
""" Geometry classes, for manipulating points, collections of points, lines, and normalizing angles. """
from math import sin, cos, pi, sqrt, atan2


def clip(value, m1, m2):
    """ Clip a value between two bounds.

    Args:
        value (float): The value to clip.
        m1 (float): The first bound.
        m2 (float): The second bound.

    Returns:
        float: A clipped value guaranteed to be between the min and max of the bounds.
    """
    lower = min(m1, m2)
    upper = max(m1, m2)
    if value > upper:
        return upper
    elif value < lower:
        return lower
    else:
        return value


def ccw(p1, p2, p3):
    """ Determine whether the turn formed by points p1, p2, and p3 is counterclockwise.

    Args:
        p1: An `(x, y)` tuple or `Point` as the start of the turn.
        p2: An `(x, y)` tuple or `Point` as the midpoint of the turn.
        p3: An `(x, y)` tuple or `Point` as the end of the turn.
    """
    return (p2[0]-p1[0]) * (p3[1]-p1[1]) > (p2[1]-p1[1]) * (p3[0]-p1[0])


def normalize_angle_180(theta):
    """ Normalize an angle in radians to be within `-pi` and `pi`.

    Args:
        theta (float): The angle to normalize, in radians.

    Returns:
        float: The normalized angle.
    """
    return ((theta+pi) % (2*pi)) - pi


def normalize_angle_360(theta):
    """ Normalize an angle in radians to be within `0` and `2*pi`.

    Args:
        theta (float): The angle to normalize, in radians.

    Returns:
        float: The normalized angle.
    """
    return theta % 2*pi


class Point:
    """ Represents a point in the x, y plane.

    Points can be interpreted and treated as x, y tuples in most cases.

    Attributes:
        x (float): The x coordinate of the point.
        y (float): The y coordinate of the point.

    Args:
        x (float): The x coordinate of the point.
        y (float): The y coordinate of the point.
    """
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ')'

    def __getitem__(self, val):  # This is so we can cheat and use xy tuples as 'other' inputs
        return self.xy_tuple()[val]

    def xy_tuple(self):
        """ Returns: An `(x, y)` tuple representing the point. """
        return self.x, self.y

    def scale(self, value, other=(0, 0)):
        """ Scale the vector from other to self by value, and return the endpoint of that vector.

        Args:
            value (float): The value to scale by.
            other: An `(x, y)` tuple or a  `Point` as the origin to scale from.

        Returns:
            The endpoint of the scaled vector, as a `Point`.
        """
        d_x, d_y = self.x-other[0], self.y-other[1]
        return Point(d_x*value+other[0], d_y*value+other[1])

    def add(self, other):
        """ Vector addition; adds two points.

        Args:
            other: An `(x, y)` tuple or an instance of `Point`.

        Returns:
            The `Point` that is the sum of this point and the argument.
        """
        return Point(self.x+other[0], self.y+other[1])

    def sub(self, other):
        """ Vector subtraction; subtracts subtracts two points.

        Args:
            other: An `(x, y)` tuple or an instance of `Point`.

        Returns:
            The `Point` that is the difference of this point and the argument.
        """
        return Point(self.x-other[0], self.y-other[1])

    def rotate(self, other, theta):
        """ Rotate about other by theta radians (positive values are counterclockwise).

        Args:
            other: An `(x, y)` tuple or an instance of `Point`.
            theta (float): The number of radians to rotate counterclockwise.

        Returns:
            The rotated `Point`.
        """
        x, y = self.x, self.y
        p_x, p_y = other[0], other[1]
        c, s = cos(theta), sin(theta)
        return Point((x-p_x)*c-(y-p_y)*s+p_x, (x-p_x)*s+(y-p_y)*c+p_y)

    def midpoint(self, other):
        """ Return a new `Point` that is the midpoint of self and other.

        Args:
            other: An `(x, y)` tuple or an instance of `Point`.

        Returns:
            A `Point` that is the midpoint of self and other.
        """
        return Point((self.x+other[0])/2.0, (self.y+other[1])/2.0)

    def distance(self, other):
        """ Calculate the distance between two points.

        Args:
            other: An `(x, y)` tuple or an instance of `Point`.

        Returns:
            float: The Euclidean distance between the points.
        """
        return sqrt(abs(self[0]-other[0])**2+abs(self[1]-other[1])**2)

    def is_near(self, other, eps):
        """ Determine whether the distance between two points is within a certain value.

        Args:
            other: An `(x, y)` tuple or an instance of `Point`.
            eps (float): The epilson within which to consider the points near one another.

        Returns:
            bool: `True` if the points are withing `eps` of each other, `False` otherwise.
        """
        return self.distance(other) < eps

    def magnitude(self):
        """ The magnitude of this point interpreted as a vector.

        Returns:
            float: The magnitude of the vector from the origin to this point.
        """
        return sqrt(self[0]**2+self[1]**2)

    def angle_to(self, other):
        """ Return the angle between two points.

        Args:
            other: An `(x, y)` tuple or an instance of `Point`.

        Returns:
            float: Angle in radians of the vector from self to other.
        """
        d_x = other[0]-self[0]
        d_y = other[1]-self[1]
        return atan2(d_y, d_x)

    def copy(self):
        """ Returns a copy of the `Point`. """
        return Point(*self)


class Pose(Point):
    """ A point facing a direction in the xy plane.

    Poses can be interpreted and treated as `(x, y, theta)` tuples in most cases.

    Args:
        x: The x coordinate of the pose.
        y: The y coordinate of the pose.
        t: The angle between the direction the pose is facing and the positive x axis, in radians.
    """

    def __init__(self, x, y, t):
        Point.__init__(self, x, y)
        self.t = t

    def __str__(self):
        return '(' + str(self.x) + ', ' + str(self.y) + ', ' + str(self.t) + ')'

    def __getitem__(self, val):  # This is so we can cheat and use xyt tuples as 'other' inputs
        return self.xyt_tuple()[val]

    def xyt_tuple(self):
        """ Returns: An `(x, y, t)` tuple representing the pose. """
        return self.x, self.y, self.t

    def transform(self, other):
        """ Return a new pose that has been transformed (translated and turned).

         Args:
             other: A `Pose` or 3-tuple-like object, by which to translate and turn.

         Returns:
             A new `Pose` equivalent to translating `self` by `(other[0], other[1])` and rotating by
             `other[2]`.
         """
        return Pose(self.x+other[0], self.y+other[1], (self.t+other[2]) % (2.0*pi))

    def rotate(self, pivot, theta):
        """ Rotate the point portion of the pose about a pivot. Leaves the theta portion unchanged.

        Args:
            pivot: A `Point`, subclass (like `Pose`), or `(x, y)` tuple as the pivot/axis of rotation.
            theta (float): The number of radians to rotate by. Positive values are counterclockwise.
        """
        x, y = Point.rotate(self, pivot, theta)
        return Pose(x, y, self.t)

    def is_near(self, other, dist_eps, angle_eps):
        """ Determine whether two poses are close.

        Args:
            other: An `(x, y, t)` tuple or  `Pose`.
            dist_eps (float): The distance epilson within which to consider the poses close.
            angle_eps (float): The angle episilon within which to consider the poses close.

        Returns:
            bool: `True` if the distance between the point portions is within `dist_eps`, and the normalized difference
            between the angle portions is within `angle_eps`.
        """
        return Point.is_near(self, other, dist_eps) and abs(normalize_angle_180(self[2]-other[2])) < angle_eps

    def point(self):
        """ Strips the angle component of the pose and returns a point at the same position.

        Returns:
            The `(x, y)` portion of the pose, as a `Point`.
        """
        return Point(self.x, self.y)

    def copy(self):
        """ Returns a copy of the Pose. """
        return Pose(*self)


class PointCollection:
    """ A movable collection of points.

    Can be iterated over like a list of `Point`. Unlike `Point`, PointCollections are mutable--that is, rotating them,
    translating them, etc. changes the internal point list.

    Args:
        points: A list of `(x, y)` tuples or `Point`.
        center: An `(x, y)` tuple or `Point` as the pivot or center of the collection.
    """
    def __init__(self, points, center=None):
        self.points = []
        for p in points:
            if isinstance(p, Point):
                self.points.append(p)
            else:
                self.points.append(Point(p[0], p[1]))
        l = len(self.points)
        if center is None:
            self.center = Point(sum([t[0] for t in self.points])/l, sum([t[1] for t in self.points])/l)
        else:
            self.center = Point(*center)

    def __getitem__(self, item):
        return self.points[item]

    def __len__(self):
        return len(self.points)

    def __str__(self):
        return str([str(p) for p in self.points])

    def scale(self, value, origin=(0, 0)):
        """ Scale each point away/towards some origin.

        Args:
            value (float): The scale amount.
            origin: An `(x, y)` tuple or `Point` from which the collection's points will move away/towards.
        """
        self.points = [p.scale(value, origin) for p in self.points]
        self.center = self.center.scale(value, origin)

    def translate(self, delta):
        """ Translate the collection by the vector delta.

        Args:
            delta: An `(x, y)` tuple or `Point`, treated as a vector and added to each point in the collection.
        """
        self.points = [p.add(delta) for p in self.points]
        self.center = self.center.add(delta)

    def rotate(self, pivot, theta):
        """ Rotate about other by theta radians (positive values are counterclockwise).

        Args:
            pivot: An `(x, y)` tuple or a `Point`.
            theta (float): The number of radians to rotate counterclockwise.
        """
        self.points = [p.rotate(pivot, theta) for p in self.points]
        self.center = self.center.rotate(pivot, theta)

    def recenter(self, new_center):
        """ Re-center the collection.

        Args:
            new_center: An `(x, y)` tuple or `Point` that will be the collection's new center.
        """
        diff = Point(new_center[0], new_center[1]).sub(self.center)
        self.translate(diff)


class Line:
    """ A line in the `(x, y)` plane defined by two points on the line.

    Args:
        p1: An `(x, y)` tuple or `Point` as one of the points on the line.
        p2: An `(x, y)` tuple or `Point` as another point on the line.
        eps (float, optional): The epsilon within which to consider a line horizontal or vertical, for precision
            purposes.
        normalize (bool, optional): If `True`, normalize the internal vector representation to be a unit vector.
    """
    def __init__(self, p1, p2, eps=1e-8, normalize=False):
        self.p1 = p1
        self.p2 = p2
        x0, y0 = p1
        x1, y1 = p2
        if abs(x1-x0) < eps:
            self.a = 1.0
            self.b = 0
            self.c = x0
        elif abs(y1-y0) < eps:
            self.a = 0.0
            self.b = 1.0
            self.c = y0
        else:
            d_x = x1-x0
            d_y = y1-y0
            if normalize:
                norm = sqrt(d_x**2+d_y**2)
                self.a = d_y/norm
                self.b = -d_x/norm
            else:
                self.a = d_y
                self.b = -d_x
            self.c = self.a*x0+self.b*y0

    def distance_from_line(self, p):
        """ Determine the (signed) distance of a point from the line.

        Args:
            p: An `(x, y)` tuple or `Point` to measure distance from.

        Returns:
            float: The signed distance from the line.
        """
        return self.a*p[0]+self.b*p[1] - self.c

    def has_point(self, p, eps=1e-8):
        """ Determine whether a point lies on the line.

        Args:
            p: The `(x, y)` tuple or `Point` to check.
            eps (float, optional): The distance to tolerate before a point is considered not to be on the line.

        Returns:
            bool: `True` if the point is on the line, `False` otherwise.
        """
        return abs(self.distance_from_line(p)) < eps

    def intersection(self, other, eps=1e-8):
        """ Determine whether two lines intersect.

        Args:
            other: The `Line` to find the intersection with.
            eps (float, optional): The smallest absolute difference to tolerate before the lines are considered to be
                converging.

        Returns:
            The `Point` of intersection, or `None` if the lines are parallel (based on epsilon).
        """
        det = self.a*other.b-other.a*self.b
        if abs(det) < eps:
            return None  # The lines are parallel, or almost parallel
        x = (other.b*self.c - self.b*other.c)/det
        y = (self.a*other.c - other.a*self.c)/det
        return Point(x, y)


class LineSegment(Line):
    """ A line segment in the `(x, y)` plane defined by two endpoints.

    Args:
        p1: An `(x, y)` tuple or `Point` as the first endpoint.
        p2: An `(x, y)` tuple or `Point` as the second endpoint.
        eps (float, optional): The minimum absolute difference in x or y before the line is considered horizontal or
            vertical.
    """
    def __init__(self, p1, p2, eps=1e-8):
        Line.__init__(self, p1, p2, eps)

    def has_point(self, p, eps=1e-8):
        """ Determine whether a point lies on the line segment.

        Args:
            p: The `(x, y)` tuple or `Point` to check.
            eps (float, optional): The distance to tolerate before a point is considered not to be on the line
                segment.

        Returns:
            bool: `True` if the point is on the line segment, `False` otherwise.
        """
        if not Line.has_point(self, p, eps):
            return False
        x, y = p[0], p[1]
        min_x, min_y = min(self.p1[0], self.p2[0]), min(self.p1[1], self.p2[1])
        max_x, max_y = max(self.p1[0], self.p2[0]), max(self.p1[1], self.p2[1])
        if abs(max_x-min_x) < eps:  # The line is vertical
            return abs(max_x-x) < eps and min_y <= y <= max_y
        elif abs(max_y - min_y) < eps:  # The line is horizontal
            return abs(max_y - y) < eps and min_x <= x <= max_x
        else:
            return min_x <= x <= max_x and min_y <= y <= max_y

    def has_intersect(self, other):
        """ Determine whether one segment intersects with another.

        Args:
            other: Another `LineSegment`.

        Returns:
            bool: `True` if the segments have an intersection, `False` otherwise.
        """
        if isinstance(other, LineSegment):  # We only check for intersections with other segments
            a1, b1 = self.p1, self.p2
            a2, b2 = other.p1, other.p2
            return ccw(a1, b1, a2) != ccw(a1, b1, b2) and ccw(a2, b2, a1) != ccw(a2, b2, b1)
        else:
            return False

    def intersection(self, other, eps=1e-8):
        """ Find the intersection(s) between two line segments, or this line segment and a `Line`.

        Args:
            other: The other `LineSegment` to find intersections with.
            eps (float, optional): The epsilon or tolerance to pass to the `Line` intersection and `has_point` checks.

        Returns:
            Either a list of `Point` (s) representing all of the intersections, or `None`, if there weren't any. Also
            returns `None` if the segment and a `Line` are exactly parallel.
        """
        if isinstance(other, LineSegment):  # Intersection between two segments
            if not self.has_intersect(other):  # Check whether the segments intersect in the first place
                return None
            potential = Line.intersection(self, other, eps)
            if potential is None:  # The segments are parallel, so find the endpoints they share
                potential = [self.p1, self.p2, other.p1, other.p2]
                potential = list(set(filter(lambda p: self.has_point(p, eps) and other.has_point(p, eps), potential)))
                return potential if len(potential) > 0 else None
            # Otherwise, just check whether the potential point is on both segments
            return [potential] if self.has_point(potential, eps) and other.has_point(potential, eps) else None
        # Otherwise, intersection between a segment (self) and a line (other)
        potential = Line.intersection(self, other, eps)
        return [potential] if potential is not None and self.has_point(potential, eps) else None



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/hooks.py
""" Functions for brains to hook into various elements of Soar.

To make use of a given hook, a brain should import it from here. It will be redefined by the client or controller
accordingly.

Treat all hook names as reserved words, and do not use them as arbitrary variable names.
"""


def tkinter_hook(widget, linked=True):  # TODO: Deprecate or merge with tkinter_wrap in 2.0
    """ Hook a Tkinter widget created by a brain into Soar, so that the UI is aware of it.

    This function is redefined by the client before a brain is ever loaded.

    Brains that import this hook can expect that their methods will always run in the main thread when running in GUI
    mode, as Tkinter is not thread-safe. Otherwise, this is not guaranteed.

    If not running in GUI mode, importing and using this hook simply returns its argument unchanged and does nothing.

    Args:
        widget: The Tkinter widget to attach to Soar. This may also be some object that supports a `destroy()` method.
        linked (bool): If `True`, the widget will be destroyed whenever the controller reloads.

    Returns:
        The window or object that was attached.
    """
    return widget


def tkinter_wrap(widget, linked=True):
    """ Make Soar's UI aware of a Tk widget, and wrap its callables so that they run on the main thread.

    This function is redefined by the client before a brain is ever loaded.
    
    This is a 'softer' version of `tkinter_hook`--it does not force all brain methods to run on the main thread.

    Rather, it attempts to find and wrap widget callables so that, when called, they will execute on the main thread.

    If not running in GUI mode, importing and using this hook simply returns its argument unchanged.

    Args:
        widget: The Tk widget to attach to Soar. This should support a `destroy()` method.
        linked (bool): If `True`, the widget will be destroyed whenever the controller reloads.

    Returns:
        The widget, with its callables wrapped so that they will run on the main thread.
    """
    return widget


def is_gui():
    """ Return whether Soar is running in GUI mode.

    Set by the client when it first loads.

    Returns:
        bool: `True` if running in GUI mode, `False` if headless.
    """
    return False


def sim_completed(obj=None):
    """ Called by the brain to signal that the simulation has completed.

    Set by the controller before the brain's `on_load()` function is called.

    Args:
        obj (optional): Optionally, an object to log to the logfile after the simulation has completed.
    """
    pass


def elapsed():
    """ Get the time that has elapsed running the controller.

    Set by the controller before the brain's `on_load()` function is called.

    Returns:
        float: The elapsed time in seconds, as defined in :attr:`soar.controller.Controller.elapsed`.
    """
    return 0.0


def raw_print(*args, **kwargs):  # TODO: Deprecate in favor of argument to print in 2.0.
    """ Allows a brain to print without the `'>>>'` prepended.

    All arguments and keyword arguments are passed to `print()`.
    """
    return print(*args, **kwargs)






# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/sim/world.py
""" Soar World and WorldObject classes/subclasses, for simulating and drawing worlds. """
import random
from math import sin, cos, pi

from soar.sim.geometry import Point, LineSegment, PointCollection


class WorldObject:
    """ An object that can be simulated and drawn in a :class:`soar.sim.world.World` on a
    :class:`soar.gui.canvas.SoarCanvas`.

    Classes that are designed to work with a World in Soar may either subclass from this class or implement its methods
    and attributes to be considered valid.

    Attributes:
        do_draw (bool): Used by a `World` to decide whether to draw the object on a canvas.
        do_step (bool): Used by a `World` to decide whether to step the object in simulation.

    Args:
        do_draw (bool): Sets the value of the `do_draw` attribute.
        do_step (bool): Sets the value of the `do_step` attribute.
        dummy (bool): Whether the object is a dummy--that is, not intended to be drawn or stepped, but used for
            some intermediate calcuation (usually collision).
        **options: Tkinter options. This may include 'tags', for drawing on a canvas, line thickness, etc.
    """
    _unique_id = 0

    def __init__(self, do_draw, do_step, dummy=False, **options):  # TODO: Replace do_draw, do_step with static, dummy, etc in 2.0
        self.do_draw = do_draw
        self.do_step = do_step
        self.dummy = dummy
        self.options = options
        if dummy:
            self.tags = None
        elif 'tags' in options:
            self.tags = options['tags']
        else:  # If tags are not specified, make sure this object has a unique tag
            self.tags = self.__class__.__name__ + str(self.__class__._unique_id)
            self.__class__._unique_id += 1
        self.options['tags'] = self.tags

    def draw(self, canvas):
        """ Draw the object on a canvas.

        Args:
            canvas: A Tkinter Canvas or a subclass, typically a :class:`soar.gui.canvas.SoarCanvas`,
                on which the object will be drawn.
        """
        pass

    def delete(self, canvas):  # TODO: Deprecate this in 2.0
        """ Delete the object from a canvas.

        Args:
            canvas: A Tkinter Canvas or a subclass, typically a :class:`soar.gui.canvas.SoarCanvas`, from which the
                object will be deleted.
        """
        if not self.dummy:
            canvas.delete(self.tags)

    def on_step(self, step_duration):
        """ Simulate the object for a step of a specified duration.

        Args:
            step_duration: The duration of the step, in seconds.
        """
        pass

    def collision(self, other, eps=1e-8):
        """ Determine whether two `WorldObject` (s) collide.

        Objects that subclass `WorldObject` should implement collision detection for every applicable class from which
        they inherit.

        Args:
            other: A supported `WorldObject` subclass with which this object could potentially collide.
            eps (float, optional): The epsilon within which to consider a collision to have occurred, different for
                each subclass.

        Returns:
            list: A list of `(x, y)` tuples consisting of all the collision points with `other`, or `None`
            if there weren't any.
        """
        pass


class Wall(WorldObject, LineSegment):
    """ A straight wall with Tk options and collision detection.

    Note that these walls are infinitely thin, so on-edge collision cannot occur.

    Args:
        p1: An `(x, y)` tuple or a :class:`soar.sim.geometry.Point` as the first endpoint of the wall.
        p1: An `(x, y)` tuple or a :class:`soar.sim.geometry.Point` as the second endpoint of the wall.
        eps (float): The epsilon within which to consider a wall vertical or horizontal.
        **options: Tkinter options.
    """
    def __init__(self, p1, p2, eps=1e-8, **options):
        WorldObject.__init__(self, do_draw=True, do_step=False, **options)
        LineSegment.__init__(self, p1, p2, eps)
        if 'width' not in options:
            self.options['width'] = 2.0  # By default walls are 2 pixels wide

    def draw(self, canvas):
        """ Draw the object on a canvas.

        Args:
            canvas: A Tkinter Canvas or a subclass, typically a :class:`soar.gui.canvas.SoarCanvas`, on which the
                object will be drawn.
        """
        if not self.dummy:
            canvas.create_line(self.p1[0], self.p1[1], self.p2[0], self.p2[1], **self.options)
            self.do_draw = False  # For a line drawn each frame, subclass this class

    def collision(self, other, eps=1e-8):
        """ Determine whether two walls intersect.

        Args:
            other: A `Wall`, or other `LineSegment` subclass.
            eps (float): The epsilon within which to consider two parallel lines the same line.

        Returns:
            A list of `(x, y)` tuples consisting of the intersection(s), or `None` if the segments do not intersect.
        """
        if isinstance(other, LineSegment):
            intersects = LineSegment.intersection(self, other, eps)
            return intersects


class Ray(Wall):
    """ A ray of a specified length, origin and direction.

    Args:
        pose: An `(x, y, theta)` tuple or :class:`soar.sim.geometry.Pose` as the origin of the ray.
        length (float): The length of the ray.
        **options: Tkinter options.
    """
    def __init__(self, pose, length, eps=1e-8, **options):
        x0, y0, theta = pose
        x1, y1 = x0+cos(theta)*length, y0+sin(theta)*length
        Wall.__init__(self, (x0, y0), (x1, y1), eps=eps, **options)
        self.length = length


class Polygon(PointCollection, WorldObject):
    """ A movable polygon with Tkinter options.

    Args:
        points: A list of `(x, y)` tuples or :class:`soar.sim.geometry.Point`.
        center: An `(x, y)` tuple or :class:`soar.sim.geometry.Point` as the pivot or center of the collection.
        **options: Tkinter options.
    """
    def __init__(self, points, center=None, **options):
        PointCollection.__init__(self, points, center)
        WorldObject.__init__(self, do_draw=True, do_step=False, **options)

    @property
    def borders(self):  # Build perimeter lines for collision detection
        return [LineSegment(self.points[i-1], self.points[i]) for i in range(len(self.points))]

    def draw(self, canvas):
        """ Draw the object on a canvas.

        Args:
            canvas: A Tkinter Canvas or a subclass, typically a :class:`soar.gui.canvas.SoarCanvas`, on which the
                object will be drawn.
        """
        if not self.dummy:
            flat_points = [p for pair in self.points for p in pair]
            canvas.create_polygon(*flat_points, **self.options)
            self.do_draw = False

    def collision(self, other, eps=1e-8):
        """ Determine whether the polygon intersects with another `WorldObject`.

        Args:
            other: Either a `Polygon` or a `Wall` as the other object.
            eps (float, optional): The epsilon within which to consider a collision to have occurred.
        """
        if isinstance(other, Polygon):  # Might be able to do this better I suppose
            intersects = []
            for i in self.borders:
                for j in other.borders:
                    line_intersects = i.intersection(j, eps=eps)
                    if line_intersects:
                        intersects.extend(line_intersects)
            return intersects if len(intersects) > 0 else None
        elif isinstance(other, Wall):
            intersects = []
            for i in self.borders:
                line_intersects = i.intersection(other, eps=eps)
                if line_intersects:
                    intersects.extend(line_intersects)
            return intersects if len(intersects) > 0 else None


class Block(Polygon):
    """ An arbitrarily thick wall centered on a line.

    Useful when infinitely-thin lines are causing issues with collision detection.

    Args:
        p1: An `(x, y)` tuple or :class:`soar.sim.geometry.Point` as the first endpoint of the line segment.
        p1: An `(x, y)` tuple or :class:`soar.sim.geometry.Point` as the second endpoint of the line segment.
        thickness (float): The thickness of the block to expand out from the line on which it is centered.
        **options: Tkinter options.
    """
    def __init__(self, p1, p2, thickness=0.002, **options):
        self.thickness = thickness
        # Build the perimeter of the wall as points
        p1 = Point(*p1)
        p2 = Point(*p2)
        mid = p1.midpoint(p2)
        points = []
        for endpoint in [p1, p2]:
            temp = endpoint.scale(1.0+thickness/endpoint.distance(mid), mid)
            pivot = temp.midpoint(endpoint)
            temp2 = temp.rotate(pivot, -pi / 2)
            temp = temp.rotate(pivot, pi/2)
            points.extend([temp, temp2])
        Polygon.__init__(self, points, center=None, **options)
        if 'fill' not in options:
            options.update({'fill': 'black'})

    @property
    def borders(self):  # Build perimeter lines dynamically
        return [Wall(self.points[i-1], self.points[i], **self.options) for i in range(len(self.points))]

    def draw(self, canvas):
        """ Draw the object on a canvas.

        Args:
            canvas:  A Tkinter Canvas or a subclass, typically a :class:`soar.gui.canvas.SoarCanvas`, on which the
                object will be drawn.
        """
        if not self.dummy:
            for wall in self.borders:
                wall.draw(canvas)
            self.do_draw = False


class World:
    """ A simulated world containing objects that can be simulated stepwise and drawn on a
    :class:`soar.gui.canvas.SoarCanvas`.

    Attributes:
        dimensions (tuple): An `(x, y)` tuple representing the worlds length and height.
        initial_position: An `(x, y, theta)` or :class:`soar.sim.geometry.Pose` representing the robot's
                          initial position in the world.
        objects (list): A list of (`WorldObject`, layer) tuples containing all of the world's objects.
        layer_max (int): The highest layer currently allocated to an object in the world.
        canvas: An instance of :class:`soar.gui.canvas.SoarCanvas`, if the world is being drawn, otherwise `None`.

    Args:
        dimensions (tuple): An `(x, y)` tuple representing the worlds length and height.
        initial_position: An `(x, y, theta)` or :class:`soar.sim.geometry.Pose` representing the robot's
                          initial position in the world.
        objects (list): The initial `WorldObject` (s) to add to the world
    """
    def __init__(self, dimensions, initial_position, objects=None):
        self.dimensions = dimensions
        self.initial_position = initial_position
        self.objects = []
        self.layer_max = -1
        self.canvas = None
        if objects:
            for obj in objects:
                self.add(obj)
        # Build boundary walls
        x, y = self.dimensions
        for wall in [Wall((0, 0), (x, 0)), Wall((x, 0), (x, y)), Wall((x, y), (0, y)), Wall((0, y), (0, 0))]:
            self.add(wall)

    def __getitem__(self, item):
        """ Iterating over a world is the same as iterating over the (sorted) object list. """
        return self.objects[item][0]

    def add(self, obj, layer=None):
        """ Add an object to the world, with an optional layer specification.

        Args:
            obj: A `WorldObject` (or a subclass instance).
            layer (int): The layer on which the object is to be drawn. Objects are drawn in order from smallest to
                largest layer. If this argument is `None`, the object's layer will be set to one higher than the
                highest layer in the objects list.
        """
        if layer is None:
            layer = self.layer_max + 1
            self.layer_max += 1
        elif layer > self.layer_max:
            self.layer_max = layer
        self.objects.append((obj, layer))
        self.objects.sort(key=lambda tup: tup[1])  # Sort the list of objects by layer priority
        setattr(obj, 'world', self)  # Ensure that every object has a back reference to the world

    def draw(self, canvas):
        """ Draw the world on a canvas.

        Objects are drawn in order from the lowest to highest layer if their `do_draw` attribute is True.

        Args:
            canvas: The :class:`soar.gui.canvas.SoarCanvas` on which to draw the world. How each object is drawn is up
                to the object.
        """
        self.canvas = canvas
        for obj, layer in self.objects:  # The list of objects is already sorted by layer
            if obj.do_draw:
                obj.draw(canvas)

    def delete(self, canvas):  # TODO: Deprecate this in 2.0
        """ Delete the world from a canvas, by deleting each object at a time.

        Args:
            canvas: The :class:`soar.gui.canvas.SoarCanvas` from which to delete.
        """
        for obj, layer in self.objects:
            if obj.do_draw:  # Objects only need to be deleted if they were drawn
                obj.delete(canvas)

    def on_step(self, step_duration):
        """ Perform a single step on the world's objects.

        Args:
            step_duration (float): The duration of the step in seconds.
        """
        for obj, layer in self.objects:
            if obj.do_step:
                obj.on_step(step_duration)

    def find_all_collisions(self, obj, eps=1e-8, condition=None):
        """ Finds all the collisions of a `WorldObject` subclass with objects in the world.

        Args:
            obj: A `WorldObject` or subclass instance. Objects in the world must know how to collide with it.
            eps (float, optional): An optional epsilon within which to consider a collision to have occurred. What that
                means differs between `WorldObject` subclasses.
            condition (optional): A function to apply to each object in the world that must be `True` in order for it
                to be considered.

        Returns:
            list: A list of `(world_obj, p)` tuples, where `world_obj` is the object that collided and `p` is the
            :class:`soar.sim.geometry.Point` at which the collision occurred. If multiple collision points occurred with
            the same object, each will be listed separately. If no collisions occurred, returns `None`.
        """
        collisions = []
        for world_obj in self:
            if condition is None or condition(world_obj):  # True if no condition or if the object matches
                obj_collisions = world_obj.collision(obj, eps)
                if obj_collisions:
                    for single_collision in obj_collisions:
                        collisions.append((world_obj, single_collision))
        return collisions if len(collisions) > 0 else None




from math import pi

from soar.sim.world import *
world = World(dimensions=(8, 1.5), initial_position=(2.0, 0.5, -pi/6), objects=[Wall((2, 1), (2.5, 1)),
                                                                                Wall((3.25, 1), (3.75, 1)),
                                                                                Wall((4.5, 1), (5.0, 1))])



"""  Tkinter wrapper for plotting using matplotlib_

.. _matplotlib: http://matplotlib.org

Note:
    Unlike use of the :func:`soar.hooks.tkinter_hook`, use of this module will not force brain methods to run on the
    main thread alongside Soar's GUI event loop.

    However, methods of `PlotWindow` *will* run on the main thread, regardless of what thread the other brain methods
    run in.

    `PlotWindow` is wrapped by the client to ensure that the proper mode (GUI or headless) is enforced regardless
    of the arguments passed to the constructor by the brain methods.

    The client will also ensure that, if logging is occurring, any `PlotWindow` objects will have their image data
    included in the log whenever the controller is shut down.

Based on code written by Adam Hartz, August 1st 2012.
"""
# Tk backends, used for plotting in various modes (GUI or headless)
import matplotlib
import platform
if platform.system() == 'Darwin':  # Fix for Soar crashing on macOS
    matplotlib.use('TkAgg')
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, FigureCanvasAgg, NavigationToolbar2TkAgg
import matplotlib.pyplot as plt
from tkinter import Tk, Toplevel, TOP, BOTH


class PlotWindow(plt.Figure):
    """ Tk window containing a matplotlib plot. In addition to the functions described below, also supports all
    functions contained in matplotlib's Axes_ and Figure_ objects.

    .. _Axes: http://matplotlib.sourceforge.net/api/axes_api.html
    .. _Figure: http://matplotlib.sourceforge.net/api/figure_api.html

    Args:
        title (str): The title to be used for the initial window.
        visible (bool): Whether to actually display a Tk window (set to `False` to create and save plots without
            displaying a window).
        standalone (bool, optional): If `True`, plot windows will be kept open by keeping the Tk event loop alive.
    """
    _tk_started = False  # If this is True, uses Toplevel to create the window, otherwise creates a main window

    def __init__(self, title="Plotting Window", visible=True):
        plt.Figure.__init__(self)
        self.add_subplot(111)
        self.visible = visible
        self._destroyed = False
        if self.visible:  # If visible, use Tk's frontend
            # Use the correct method to create a window, then set the tk_started flag to True
            self.canvas = FigureCanvasTkAgg(self, Toplevel() if self.__class__._tk_started else Tk())
            self.__class__._tk_started = True
            self.title(title)
            self.make_window()
            self.show()
        else:
            self.canvas = FigureCanvasAgg(self)

    def make_window(self):
        """ Pack the plot and matplotlib toolbar into the containing Tk window.

        This method is called during initialization and it is unlikely you will need to call it elsewhere.
        """
        self.canvas.get_tk_widget().pack(side=TOP, fill=BOTH, expand=1)
        self.toolbar = NavigationToolbar2TkAgg(self.canvas, self.canvas._master)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=TOP, fill=BOTH, expand=1)

    def destroy(self):
        """ Destroy the Tk window.

        Note that after calling this method (or manually closing the Tk window), this :py:class:`PlotWindow` cannot be
        used.
        """
        try:  # Try and destroy the window
            self.canvas._master.destroy()
        except:
            pass  # It was probably already destroyed
        self._destroyed = True

    def clear(self):
        """ Clear the plot, keeping the Tk window active. """
        self.clf()
        self.add_subplot(111)
        if self.visible:
            self.show()

    def show(self):
        """ Update the canvas image (automatically called for most functions). """
        self.canvas.show()

    def __getattr__(self, name):
        show = True
        if name.startswith('_'):
            name = name[1:]
            show = False
        if hasattr(self.axes[0], name):
            attr = getattr(self.axes[0], name)
            if hasattr(attr,'__call__'):
                if show:
                    def tmp(*args,**kwargs):
                        out = attr(*args,**kwargs)
                        if self.visible:
                            self.show()
                        return out
                    return tmp
                else:
                    return attr
            else:
                return attr
        else:
            raise AttributeError("PlotWindow object has no attribute %s" % name)
        
    def title(self, title):
        """ Change the title of the Tk window """
        self.canvas._master.title(title)
        
    def legend(self, *args):
        """ Create a legend for the figure (requires plots to have been made with labels) """
        handles, labels = self.axes[0].get_legend_handles_labels()
        self.axes[0].legend(handles, labels)
        if self.visible:
            self.show()

    def save(self, fname, **kwargs):
        """ Save this plot as an image.  File type determined by extension of filename passed in.

        See documentation for savefig_.

        .. _savefig: http://matplotlib.sourceforge.net/api/figure_api.html

        Args:
            fname: The file to create, which may be a path or a file-like object.
        """
        self.savefig(fname, **kwargs)

    def stay(self):
        """ Start the Tkinter window's main loop (e.g., to keep the plot open at the end of the execution of a script)
        """
        self.canvas._master.mainloop()

    def plot(self, *args, **kwargs):
        """ Plot lines and/or markers to the Axes. See pyplot_ for more information.

        .. _pyplot: https://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.plot
        """
        self.__getattr__('plot')(*args, **kwargs)



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/errors.py
""" Soar error classes. """
import sys


def printerr(*args, **kwargs):
    """ A wrapper around `print` to print to `sys.stderr`. """
    print(*args, file=sys.stderr, **kwargs)


class SoarError(Exception):
    """ An umbrella class for Soar exceptions.

    This is raised by the client if it encounters an error it cannot handle, such as an exception that occurs in
    headless mode.
    """
    pass


class LoggingError(SoarError):
    """ Raised when an error occurs writing to a log file. Typically, these are ignored and merely printed. """
    pass


class SoarIOError(SoarError):
    """ Raised whenever an error occurs communicating with a real robot.

    Typically, this is raised within the robot class when a connection could not be established, timed out, or invalid
    data that cannot be dealt with was received.
    """
    pass


class GUIError(SoarError):
    """ Raised when an error occurs while drawing the world or one of its objects. """
    pass



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/controller.py
""" Controller classes and functions for controlling robots, simulated or real. """
from io import StringIO
from threading import Thread
from time import sleep
from timeit import default_timer as timer

from soar.common import DRAW, CONTROLLER_COMPLETE, STEP_FINISHED, MAKE_WORLD_CANVAS, EXCEPTION


class Controller:
    """ A class for interacting with simulated or real robots.

    Initialized by the client and used to call the user defined methods of the robot and brain.

    Attributes:
        running (bool): Indicates whether the controller is currently running--that is, repeatedly stepping.
        started (bool): Indicates whether the controller has been started.
        stopped (bool): Indicates whether the controller has been stopped.
        step_count (int): The number of steps that have elapsed so far.
        elapsed (float): The number of seconds spent actually running. Unless a step takes longer than
            `step_duration`, this will typically be the `step_count` multiplied by `step_duration`. If any steps take
            longer, the additional time will also be counted here.

    Args:
        client_future: The function to call to schedule a future for the client to execute.
        gui: The currently :class:`soar.gui.soar_ui.SoarUI` instance, if any, or `None` if in headless mode.
        simulated (bool): If `True`, the controller will simulate the robot. Otherwise it will treat the robot as real.
        robot: An instance of :class:`soar.robot.base.BaseRobot` or a subclass.
        brain: The currently loaded brain module, supporting the `on_load()`, `on_start()`, `on_step()`, `on_stop()`,
            and `on_shutdown()` methods.
        realtime (bool): If `True`, stepping takes real time--that is, the controller will sleep for whatever time is
            not used running the step, until the step has taken at least `step_duration` seconds. Otherwise, no
            sleeping will occur; however the elapsed time will behave as if each step was at least `step_duration` long.
        world: An instance of :class:`soar.sim.world.World` or one of its subclasses if one is loaded, or `None`.
        step_duration (float): The duration of a single step, in seconds.
        log: A callable that accepts a `dict`-like object as an argument to log to a file, or `None`, if no logging
        is to take place.
    """
    def __init__(self, client_future, robot, brain, simulated, gui, step_duration=0.1, realtime=True, world=None,
                 log=None):
        self.client_future = client_future
        self.robot = robot
        self.brain = brain
        self.step_duration = step_duration
        self.simulated = simulated
        self.gui = gui
        self.realtime = realtime
        self.world = world
        self.log = log
        self.running = False
        self.started = False
        self.stopped = False
        self.elapsed = 0.0
        self.step_count = 0
        self._step_thread = None
        self._avg_offset = 0.0
        self._brain_log_contents = StringIO()

    def load(self):
        """ Called when the controller is loaded. """
        # Set brain print function
        def brain_print(*args, **kwargs):
            raw = kwargs.pop('raw', False)
            if raw:  # Print without the three carets '>>>'
                print(*args, **kwargs)
            else:
                print('>>>', *args, **kwargs)
            if self.log:  # If logging, ensure anything printed is logged
                print(*args, file=self._brain_log_contents, **kwargs)
        self.brain['print'] = brain_print

        if 'elapsed' in self.brain:  # Give brain read-only access to the elapsed time
            self.brain['elapsed'] = lambda: self.elapsed
        if 'sim_completed' in self.brain:  # Give brain the ability to end the simulation
            def sim_completed(obj=None):
                self.client_future(CONTROLLER_COMPLETE, obj)
                self.running = False
            self.brain['sim_completed'] = sim_completed

        self.robot.simulated = self.simulated
        if self.simulated:  # If we are simulating, we have to initialize the world
            self.robot.move(self.world.initial_position)
            self.world.add(self.robot)
            if self.gui:  # If a GUI exists, we make it try to draw the world before finishing loading
                if self.gui.synchronous_future(self.gui.make_world_canvas, self.world) == EXCEPTION:
                    return EXCEPTION
        # The robot is always loaded first, in case the brain depends on it
        self.robot.on_load()
        # TODO: Figure out how to force a brain crash
        # t = Thread(target=self.force_brain_crash, daemon=True)
        # t.start()
        self.brain['on_load']()

    # def force_brain_crash(self):
    #     sleep(5.0)
    #     class Dummy:
    #         def __getattribute__(*args, **kwargs):
    #             raise Exception
    #         def __call__(*args, **kwargs):
    #             raise Exception
    #     dummy_obj = Dummy()
    #     for key in self.brain.keys():
    #         if key == '__builtins__':
    #             for b_key in self.brain[key].keys():
    #                 self.brain[key][b_key] = dummy_obj
    #         else:
    #             self.brain[key] = dummy_obj
    #     print('Attempted to force brain crash', file=sys.stderr)

    def run(self, n=None):
        """ Runs the controller, starting it if necessary, for one or many steps, or without stopping.

        Args:
            n: If `None`, run forever, at least until stopped. If 0, start the controller, if it has not yet been
                started. Otherwise, for `n > 0`, run for that many steps.
        """
        if not self.started:
            self.robot.on_start()
            self.brain['on_start']()
            self.started = True
        if n == 0:  # Don't do any steps
            return
        if n == 1:  # If we're just doing one step, no need to start a new thread
            step_time = self.single_step()
            if step_time < self.step_duration:  # Just add self.step_duration to the elapsed time
                self.elapsed += self.step_duration
            else:  # If the step took longer than it should have, add its length to self.elapsed
                self.elapsed += step_time
            self.client_future(STEP_FINISHED)
        else:  # Otherwise let the step timer run the steps
            self._step_thread = Thread(target=lambda: self.step_thread(n=n), daemon=True)
            self._step_thread.start()

    def step_thread(self, n=None):  # If n is unspecified, run forever until stopped.
        # step_thread is typically wrapped by the client so that any exceptions that occur are made known.
        self.running = True
        while self.running and (n is None or n > 0):
            start = timer()  # Time the actual step, with sleeping included
            step_time = self.single_step()
            if step_time < self.step_duration:  # Just add self.step_duration to the elapsed time
                self.elapsed += self.step_duration
                if self.realtime:  # If running in real time, sleep accordingly
                    sleep(max(0, self.step_duration-step_time-self._avg_offset))
            else:  # If the step took longer than it should have, add its length to self.elapsed and don't sleep
                self.elapsed += step_time
            if n:  # If running for a specific number of steps, decrement
                n -= 1
            if self.realtime:  # If running in real time, try and figure out by how much the sleep was too long or short
                real_step_duration = timer() - start
                offset = real_step_duration - self.step_duration
                self._avg_offset = 0.75 * self._avg_offset + 0.25 * offset  # Exponentially weighted moving average
        if self.running:  # If the timer finished naturally, without being stopped, notify the client
            self.running = False
            self.client_future(STEP_FINISHED)

    def single_step(self):  # Undergoes a single step, returns the number of seconds the step took
        if self.log:  # Log information before the step to the log file
            self.log_step_info()
        start = timer()
        # First step the brain
        self.brain['on_step'](self.step_duration)
        # Measure how long the brain took. If it took longer, the robot and world should be stepped accordingly
        brain_duration = timer()-start
        if brain_duration > self.step_duration:
            step_duration = brain_duration
        else:
            step_duration = self.step_duration
        if self.simulated:  # If simulating, the world will handle the robot's step
            self.world.on_step(step_duration)
            if self.gui:
                self.client_future(DRAW, self.world)
        else:  # Otherwise step the robot on its own
            self.robot.on_step(step_duration)
        self.step_count += 1
        return timer()-start

    def log_step_info(self):
        """ Log information about the current step. """
        log_object = {'type': 'step', 'elapsed': self.elapsed, 'step': self.step_count, 'robot': self.robot.to_dict()}
        if self._brain_log_contents.getvalue() != '':
            log_object.update({'brain_print': self._brain_log_contents.getvalue()})
            self._brain_log_contents.truncate(0)
            self._brain_log_contents.seek(0)
        self.log(log_object)

    def pause(self):  # Stops the currently running step thread, if it exists
        self.running = False
        if self._step_thread:
            self._step_thread.join()

    def stop(self):
        """ Called when the controller is stopped. """
        self.pause()
        self.brain['on_stop']()
        self.robot.on_stop()
        if self.log:  # Log position information after all steps have completed
            self.log_step_info()
        self.started = False
        self.stopped = True

    def shutdown(self):
        """ Called when the controller is shut down. """
        self.pause()
        self.brain['on_shutdown']()
        self.robot.on_shutdown()

    def failure(self):
        """ Called when the controller fails. """
        self.pause()
        self.started = False
        self.stopped = True
        try:
            self.robot.on_shutdown()  # Ignore any errors while trying to shutdown
        except Exception:
            pass



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/gui/__init__.py
""" Soar GUI Modules.

Contains the modules for building and displaying Soar's GUI, as well as all GIFs used for icons or animations, and
widgets for plot generation, etc.
"""
from soar import __version__


# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/sim/__init__.py
""" Simulation modules, containing general 2D geometry methods, graphical world objects with collision detection, etc.
"""



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/update.py
import re
from math import inf
from urllib import request

from soar import __version__


def get_update_message():
    # Try and determine if a newer version of Soar is available, and notify the user
    try:
        r = request.urlopen('https://pypi.python.org/pypi/Soar', data=None)
        assert(r.getcode() == 200)
        body = str(r.read(), encoding='utf-8')
        start = body.find('Soar')
        assert (start != -1)
        # Capture groups are Major, minor, patch, and optional .devN
        match_version = re.compile('(\d)\.(\d)\.(\d)(?:\.dev(\d))?')

        def version_parse(x):  # Parse version increment strings into integers
            if x is None:
                return inf  # Non-development releases have an effective development release value of infinity
            else:
                return int(x)

        pypi_version = list(map(version_parse, match_version.search(body, start).group(1, 2, 3, 4)))
        current_version = list(map(version_parse, match_version.search(__version__).group(1, 2, 3, 4)))
        need_update = False
        for pypi_inc, current_inc in zip(pypi_version, current_version):
            if pypi_inc > current_inc:  # If any PyPI version increment is greater, we need to update
                need_update = True
                break
            elif pypi_inc < current_inc:  # If any current version increment is greater, we do not need to update
                need_update = False
                break
        if need_update:
            if pypi_version[3] != inf:
                pypi_version[3] = 'dev' + str(pypi_version[3])
            else:
                pypi_version = pypi_version[:-1]  # Slice off the irrelevant infinite dev version
            version_string = '.'.join([str(x) for x in pypi_version])
            return 'A newer version of Soar is available: v' + version_string + '\nPlease update your installation.'
        else:
            return ''
    except Exception:  # If anything at all goes wrong, silently fail
        return ''



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# soar/__init__.py
""" Soar (Snakes on a Robot) v1.3.8

An extensible Python framework for both simulating and interacting with robots.

Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
"""
__version__ = '1.3.8'
blerb = 'Soar (Snakes on a Robot) v' + __version__ + ': A Python robotics framework.\n' \
                                                     'https://github.com/arantonitis/soar\n\n' \
                                                     'Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.'




from setuptools import setup

from soar import __version__

with open('README.rst', 'r') as f:
    long_description = f.read()

setup(name='Soar',
      version=__version__,
      description='An extensible Python framework for simulating and interacting with robots',
      long_description=long_description,
      author='Andrew Antonitis',
      author_email='andrewan@mit.edu',
      url='https://github.com/arantonitis/soar',
      packages=['soar', 'soar.brains', 'soar.gui', 'soar.robot', 'soar.sim', 'soar.worlds'],
      package_data={'': ['*.gif']},
      entry_points={'console_scripts': ['soar = soar.__main__:main',]},
      install_requires=['pyserial>=3.0', 'matplotlib>=2.0'],
      python_requires='>=3',
      license='LGPLv3',
      classifiers=['License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)'],
      )



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/robot/base.py
""" Soar BaseRobot class, intended as a parent class for nontrivial/useful robots.

All robots usable in Soar should either subclass from BaseRobot or, if this is not possible, reproduce its behaviors.
"""
from math import sin, cos

from soar.errors import SoarIOError
from soar.sim.geometry import Point, Pose
from soar.sim.world import WorldObject, Polygon, Wall


class BaseRobot(WorldObject):
    """ A base robot class, intended to be subclassed and overridden.

    Any robot usable in SoaR should supplement or re-implement this class' methods with the desired behavior.

    Attributes:
        type (str): A human readable name for robots of this type.
        world: The instance of :class:`soar.sim.world.World` (or a subclass) in which the robot resides, or `None`,
               if the robot is real.
        simulated (bool): Any BaseRobot subclass should consider the robot to be simulated if this is `True`, and real
                          otherwise. By default, it is `False`.
        pose: An instance of :class:`soar.sim.geometry.Pose` representing the robot's `(x, y, theta)` position.
            In simulation, this is the actual position; on a real robot this may be determined through other means.
        polygon: A :class:`soar.sim.world.Polygon` that defines the boundaries of the robot and is used for collision.
        fv (float): The robot's current translational velocity, in arbitrary units. Positive values indicate movement
            towards the front of the robot, and negative values indicate movement towards the back.
        rv (float): The robot's current rotational velocity, in radians/second. Positive values indicate
            counterclockwise rotation (when viewed from above), and negative values indicate clockwise rotation.

    Args:
        polygon: A :class:`soar.sim.world.Polygon` that defines the boundaries of the robot and is used for collision.
        **options: Arbitrary keyword arguments. This may include Tkinter keywords passed to the `WorldObject`
            constructor, or robot options supported as arguments to `set_robot_options`.
    """

    def __init__(self, polygon, **options):
        WorldObject.__init__(self, do_draw=True, do_step=True, **options)  # Robots are always drawn and stepped
        self.type = 'BaseRobot'
        self.simulated = True
        self.world = None
        self.pose = Pose(0, 0, 0)
        self.fv = 0.0
        self.rv = 0.0
        # Re-instantiate the polygon with the robot's tags
        del polygon.options['tags']
        self.polygon = Polygon(polygon.points, polygon.center, tags=self.tags, **polygon.options)
        # The maximum radius, used for pushing the robot back before a collision
        self._radius = sorted([self.polygon.center.distance(vertex) for vertex in self.polygon])[-1]

    def set_robot_options(self, **options):
        """ Set one or many keyworded, robot-specific options. Document these options here.

        Args:
            **options: `BaseRobot` does not support any robot options.
        """
        pass

    @property
    def pos(self):  # TODO: This property exists only for backwards compatibility. Deprecate by 2.0.
        return self.pose

    @pos.setter
    def pos(self, t):  # TODO: This property exists only for backwards compatibility. Deprecate by 2.0.
        self.pose = t

    def to_dict(self):
        """ Return a dictionary representation of the robot, usable for serialization. """
        return {'x_pos': self.pose[0], 'y_pos': self.pose[1], 't_pos': self.pose[2], 'fv': self.fv, 'rv': self.rv,
                'type': self.type}

    def move(self, pose):
        """ Move the robot to the specified `(x, y, theta)` pose.

        Args:
            pose: An :class:`soar.sim.geometry.Pose` or 3-tuple-like object to move the robot to.
        """
        x, y, t = pose
        current_theta = self.pose[2]
        self.pose = Pose(x, y, t)
        self.polygon.recenter(self.pose)
        self.polygon.rotate(self.polygon.center, t - current_theta)

    def collision(self, other, eps=1e-8):
        """ Determine whether the robot collides with an object.

        Supported objects include other robots, and subclasses of :class:`soar.sim.world.Polygon` and
        :class:`soar.sim.world.Wall`.

        Args:
            other: A supported `WorldObject` subclass with which this object could potentially collide.
            eps (float, optional): The epsilon within which to consider a collision to have occurred, different for
                each subclass.

        Returns:
            list: A list of `(x, y)` tuples consisting of all the collision points with `other`, or `None`
            if there weren't any.
        """
        if isinstance(other, BaseRobot):  # Dispatch to collisions between robot polygons
            return self.polygon.collision(other.polygon, eps=eps)
        elif isinstance(other, Polygon) or isinstance(other, Wall):
            return self.polygon.collision(other, eps=eps)

    def draw(self, canvas):
        """ Draw the robot on a canvas.

        Canvas items are preserved. If drawn more than once on the same canvas, the item will be moved and not redrawn.

        Args:
            canvas: An instance of :class:`soar.gui.canvas.SoarCanvas`.
        """
        try:  # Try and find the drawn polygon on the canvas, in case it already exists
            canvas_poly = canvas.find_withtag(self.tags)[0]
        except IndexError:  # If no such item exists, draw it for the first time
            self.polygon.draw(canvas)
        else:
            # Remap metered coordinates to pixel coordinates, and change the canvas polygon
            coords = canvas.remap_coords([p for pair in self.polygon for p in pair])
            canvas.coords(canvas_poly, coords)

    def delete(self, canvas):  # TODO: Deprecate this in 2.0
        """ Delete the robot from a canvas.

        Args:
            canvas: An instance of :class:`soar.gui.canvas.SoarCanvas`.
        """
        self.polygon.delete(canvas)

    def on_load(self):
        """ Called when the controller of the robot is loaded.

        The behavior of this method should differ depending on the value of `simulated`; if it is `False`, this
        method should be used to connect with the real robot. If a connection error occurs, a
        :class:`soar.errors.SoarIOError` should be raised to notify the client that the error was not due to other
        causes.
        """
        if not self.simulated:
            raise SoarIOError('BaseRobot has no real interface to connect to')

    def on_start(self):
        """ Called when the controller of the robot is started. 
        
        This method will always be called by the controller at most once per controller session, before the first step.
        """
        pass

    def on_pause(self):
        """ Called when the controller is paused. """
        pass

    def on_step(self, step_duration):
        """ Called when the controller of the robot undergoes a single step of a specified duration.

        For BaseRobot, this tries to perform an integrated position update based on the forward and rotational
        velocities. If the robot cannot move to the new position because there is an object in its way, it will be moved
        to a safe space just before it would have collided.

        Subclasses will typically have more complex `on_step()` methods, usually with behavior for stepping
        non-simulated robots.

        Args:
            step_duration (float): The duration of the step, in seconds.
        """
        if self.simulated:  # Do the simulated move update (with collision preemption)
            # Try and make sure that the robot can actually move to its new location
            # Integrate over the path, making the new position at the end of the arc
            theta = self.pose[2]
            d_t = self.rv*step_duration
            new_theta = theta+d_t
            if self.rv != 0:
                d_x = self.fv*(sin(new_theta)-sin(theta))/self.rv
                d_y = self.fv*(cos(theta)-cos(new_theta))/self.rv
            else:
                d_x, d_y = self.fv*cos(theta)*step_duration, self.fv*sin(theta)*step_duration
            new_pos = self.pose.transform((d_x, d_y, d_t))
            # Build a dummy wall between the old and new position and check if it collides with anything
            w = Wall(self.pose.point(), new_pos.point(), dummy=True)
            collisions = self.world.find_all_collisions(w, condition=lambda obj: obj is not self)
            if collisions:  # If there were collisions, push the robot to a safe distance from the closest one
                collisions.sort(key=lambda tup: self.pose.distance(tup[1]))
                safe_point = Point(*collisions[0][1])
                offset = Point(self._radius, 0.0).rotate((0, 0), new_pos[2])
                safe_point = safe_point.sub(offset)
                new_pos = Pose(safe_point.x, safe_point.y, new_pos[2])

            self.pose = new_pos
            self.polygon.recenter(new_pos)
            self.polygon.rotate(self.polygon.center, d_t)

    def on_stop(self):
        """ Called when the controller of the robot is stopped. """
        pass

    def on_shutdown(self):
        """ Called when the controller of the robot is shutdown.

        If interacting with a real robot, the connection should be safely closed and reset for any later connections.
        """
        pass



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/robot/__init__.py
""" Soar robot modules, containing the BaseRobot class, example robot subclasses, and bundled implementations like
PioneerRobot.
"""



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/robot/names.py
""" Generate a neutral name from an arbitrary serial number string. """
import random
from math import sqrt, ceil


def is_prime(n):
    if n < 4:
        return False if n == 1 else True
    for i in range(2, ceil(sqrt(n))):
        if n % i == 0:
            return False
    return True

names = ['Ariel', 'Bailey', 'Casey', 'Dallas', 'Eli', 'Frankie', 'Gabriel', 'Harley', 'Jayden', 'Kai', 'Lee', 'Mickey',
         'Neo', 'Ocean', 'Parris', 'Quinn', 'Reagan', 'Shiloh', 'Taylor', 'Udo', 'Val', 'Winter', 'Xue', 'Yoshi',
         'Zephyr', 'Narin', 'Ollie', 'Dakota', 'Marlie', 'Brooke', 'Shannon', 'Nour', 'Reese', 'Storm', 'Sydney',
         'Omari', 'Ola', 'Ryley', 'Esmai', 'Leslie', 'Finley', 'Hildred', 'Lynn', 'Shelby', 'Sequoia', 'Tashi',
         'Dominique', 'Garnet', 'Gray', 'Cameron']


def name_from_sernum(sernum):
    # Pull out only the digits from the serial number
    x = 0
    for y in map(lambda c: int(c), filter(lambda c: c.isdigit(), sernum)):
        x = x*10+y
    # Hash them with a Linear congruential generator
    return names[((a*x+b) % p) % m]


def test_collisions():
    collisions = 0
    for r in [range(i, i + 50) for i in range(2600, 2650 + 1)]:
        chosen = set()
        for i in r:
            name = name_from_sernum(str(i))
            if name in chosen:
                collisions += 1
            else:
                chosen.add(name)
    return collisions


def has_all():  # Test whether a hash setting contains every name for the expected serial number range
    return all([name in [name_from_sernum(str(i)) for i in range(2600, 2700)] for name in names])


m = len(names)
a = 74
b = 98
p = 149
# while test_collisions() > 25 or not has_all():
#     p = 1
#     while not is_prime(p):
#         p = random.randint(m+1, 3*m)
#     a = random.randint(2, p)
#     b = random.randint(2, p)
# print(a, b, p, test_collisions())
# for i in range(2600, 2700):
#     print(name_from_sernum(str(i)))



# Soar (Snakes on a Robot): A Python robotics framework.
# Copyright (C) 2017 Andrew Antonitis. Licensed under the LGPLv3.
#
# soar/robot/arcos.py
""" ARCOS (Advanced Robot Control and Operations Software) Client.

Classes and functions for communicating with an ARCOS server running on an Adept MobileRobot platform
(typically Pioneer 2 and 3).
"""
from threading import Thread, Lock, Event
from time import sleep

from serial import Serial, SerialTimeoutException, SerialException
from serial.tools.list_ports import comports

from soar.errors import SoarError
from soar.client import printerr

__version__ = '1.0'

# ARCOS Client command codes

SYNC0 = 0  #: The initial synchronization packet.
SYNC1 = 1  #: The second synchronization packet.
SYNC2 = 2  # The final synchronization packet. Robot specific information is sent back after this packet.
PULSE = 0  #: Reset server watchdog (typically sent every second so that the robot knows the client is alive).
OPEN = 1  #: Start the ARCOS servers.
CLOSE = 2  #: Close servers and client connection. Also stops the robot.
POLLING = 3  #: Change sonar polling sequence. Argument is a string consisting of sonar numbers 1-32 (as single bytes).
ENABLE = 4  #: Enable the motors, if argument is 1, or disable them if it is 0.
SETA = 5  #: Set translation acceleration, if positive, or deceleration, if negative, in mm/sec^2.
SETV = 6  #: Set maximum translational velocity in mm/sec. Note that the robot is still limited by the hardware cap.
SETO = 7  #: Reset local odometry position to the origin `(0, 0, 0)`.
MOVE = 8  #: Translate forward (+) or backward (-) mm absolute distance at `SETV` speed.
ROTATE = 9  #: Rotate counter- (+) or clockwise (-) degrees/sec at `SETRV` limited speed.
SETRV = 10  #: Set maximum rotation velocity in degrees/sec. Note that the robot is still limited by the hardware cap.
VEL = 11  #: Translate at mm/sec forward (+) or backward (-), capped by `SETV`.
HEAD = 12  #: Turn at `SETRV` speed to absolute heading; +-degrees (+ is counterclockwise).
DHEAD = 13  #: Turn at `SETRV` speed relative to current heading; (+) counterclockwise or (-) clockwise degrees.
SAY = 15
""" Play up to 20 duration, tone sound pairs through User Control panel piezo speaker.
The argument is a string whose first byte must be the number of (duration, tone) pairs to play, followed by each (duration, tone) pair.
Duration is in 20 millisecond increments.
A value of 0 means silence. The values 1-127 are the corresponding MIDI notes, with 60 being middle C. The remaining values are frequencies
computed as `(tone - 127)*32`, ranging from 1-4096 in 32Hz increments.
"""
CONFIG = 18  #: Request a configuration SIP.
ENCODER = 19  #: Request a single (1), a continuous stream (>1), or stop (0) encoder SIPS.
RVEL = 21  #: Rotate (degrees/sec) counterclockwise (+) or clockwise (-). Limited by `SETRV`.
DCHEAD = 22  #: Adjust heading relative to last setpoint; +- degrees (+ is counterclockwise).
SETRA = 23  #: Change rotation (+) acceleration or (-) deceleration in degrees/sec^2
SONAR = 28  #: 1=enable, 0=disable all the sonar; otherwise bits 1-3 specify an array from 1-4 to enable/disable.
STOP = 29  #: Stop the robot without disabling the motors.
DIGOUT = 30  #: Set (1) or reset (0) User Output ports. Bits 8-15 is a byte mask that selects, if set the output ports
""" Set (1) or reset (0) user output ports. High bits 8-15 is a byte mask that selects the ports to change;
low bits 0-7 set (1) or reset (0) the selected port(s).
"""
VEL2 = 32  #: Set independent wheel velocities; bits 0-7 for right wheel, bits 8-15 for left in 20mm/sec increments.
ADSEL = 35  #: Select the A/D port number for reporting ANPORT value in standard SIP.
IOREQUEST = 40  #: Request a single (1), a continuous stream (>1), or stop (0) IO SIPS.
BUMPSTALL = 44  #: Stall robot if no (0), only front (1), only rear (2), or either (3) bumpers make contact.
SONARCYCLE = 48  #: Change the sonar cycle time, in milliseconds.
E_STOP = 55  #: Emergency stop. Overrides acceleration, so is very abrupt.
SOUNDTOG = 92  #: Mute (0) or enable (1) the user control piezo speaker.

command_types = {
    PULSE: None,
    OPEN: None,
    CLOSE: None,
    POLLING: str,
    ENABLE: int,
    SETA: int,
    SETV: int,
    SETO: None,
    MOVE: int,
    ROTATE: int,
    SETRV: int,
    VEL: int,
    HEAD: int,
    DHEAD: int,
    SAY: str,
    CONFIG: None,
    ENCODER: int,
    RVEL: int,
    DCHEAD: int,
    SETRA: int,
    SONAR: int,
    STOP: None,
    DIGOUT: int,
    VEL2: int,
    ADSEL: int,
    IOREQUEST: int,
    BUMPSTALL: int,
    SONARCYCLE: int,
    E_STOP: None,
    SOUNDTOG: int,
}
""" The argument type of every supported ARCOS command. """


class ARCOSError(SoarError):
    """ Umbrella class for ARCOS-related exceptions. """


class Timeout(ARCOSError):
    """ Raised when no packet is read after a certain interval. """


class InvalidPacket(ARCOSError):
    """ Raised when a packet's checksum is incorrect. """


def packet_checksum(data):
    """ Calculate and returns the ARCOS packet checksum of a packet which does not have one.

    Args:
        data (list): A list of data bytes.

    Returns:
        int: The packet checksum.
    """
    checksum = 0
    i = 3
    n = data[2]-2
    while n > 1:
        checksum += (data[i] << 8) | data[i+1]
        checksum &= 0xffff
        n -= 2
        i += 2
    if n > 0:
        checksum ^= data[i]
    return checksum


def __b_2_i(l, i):  # Takes a list and an index and returns the two bytes combined into an int
    return l[i] | (l[i + 1] << 8)


def __str_from_i(l, i):  # Takes a list and an index and returns a string and the index after the null terminator
    s = ''
    while l[i] != 0:
        s += chr(l[i])
        i += 1
    i += 1
    return s, i


def __unpack_byte_fields(data, packet, i, *fields):  # Unpack an arbitrary number of byte fields starting at index i
    for field in fields:
        data.update({field: packet[i]})
        i += 1
    return i


def __unpack_int_fields(data, packet, i, *fields):  # Unpack an arbitrary number of int fields starting at index i
    for field in fields:
        data.update({field: __b_2_i(packet, i)})
        i += 2
    return i


def __unpack_str_fields(data, packet, i, *fields):  # Unpack an arbitrary number of str fields starting at index i
    for field in fields:
        s, i = __str_from_i(packet, i)
        data.update({field: s})
    return i


def decode_packet(packet):
    """ Decode a SIP (Server Information Packet) into a field-indexable dictionary.

    Returns:
        dict: A dictionary with field names as keys and values as corresponding numbers. The `'TYPE'` key holds a value
        of `'STANDARD'`, `'CONFIG'`, `'ENCODER'`, or `'IO'`, corresponding to the packet type.

    Raises:
        `InvalidPacket`: If a packet's fields could not be decoded.
    """
    try:
        data = {'TYPE': packet[3], 'CHECKSUM': (packet[-1] & 0xff) | (packet[-2] << 8)}
        if data['TYPE'] in [0x32, 0x33]:  # Standard sip
            data['TYPE'] = 'STANDARD'
            __unpack_int_fields(data, packet, 4, 'XPOS', 'YPOS', 'THPOS', 'L VEL', 'R VEL')
            data['BATTERY'] = packet[14]
            __unpack_int_fields(data, packet, 15, 'STALL AND BUMPERS', 'CONTROL', 'FLAGS')
            data.update({'COMPASS': packet[21], 'SONAR_COUNT': packet[22]})
            sonars = {}
            i = 23
            for sonar in range(data['SONAR_COUNT']):
                number = packet[i]
                dist = __b_2_i(packet, i + 1)
                sonars.update({number: dist})
                i += 3
            data.update({'SONARS': sonars})
            i = __unpack_byte_fields(data, packet, i, 'GRIP_STATE', 'ANPORT', 'ANALOG', 'DIGIN', 'DIGOUT')
            data.update({'BATTERYX10': __b_2_i(packet, i)})
            i += 2
            data['CHARGE_STATE'] = packet[i]
            i += 1
            data['ROTVEL'] = __b_2_i(packet, i)
        elif data['TYPE'] == 0x20:  # CONFIGpac
            data['TYPE'] = 'CONFIG'
            i = __unpack_str_fields(data, packet, 4, 'ROBOT_TYPE', 'SUBTYPE', 'SERNUM')
            data.update({'4MOTS': packet[i]})
            i = __unpack_int_fields(data, packet, i+1, 'ROTVELTOP', 'TRANSVELTOP', 'ROTACCTOP', 'TRANSACCTOP', 'PWMMAX')
            s, i = __str_from_i(packet, i)
            data.update({'NAME': s})
            i = __unpack_byte_fields(data, packet, i, 'SIPCycle', 'HOSTBAUD', 'AUXBAUD')
            i = __unpack_int_fields(data, packet, i, 'GRIPPER', 'FRONT_SONAR')
            data.update({'REAR_SONAR': packet[i]})
            i = __unpack_int_fields(data, packet, i+1, 'LOWBATTERY', 'REVCOUNT', 'WATCHDOG')
            data.update({'P2MPACS': packet[i]})
            i = __unpack_int_fields(data, packet, i+1, 'STALLVAL', 'STALLCOUNT', 'JOYVEL', 'JORVEL', 'ROTVELMAX',
                                    'TRANSVELMAX')
        elif data['TYPE'] == 0x90:  # ENCODERpac
            data['TYPE'] = 'ENCODER'
            data.update({'L_ENCODER': (__b_2_i(packet, 6) << 16) | __b_2_i(packet, 4),
                         'R_ENCODER': (__b_2_i(packet, 10) << 16) | __b_2_i(packet, 8)})
        elif data['TYPE'] == 0xF0:  # IOpac
            data['TYPE'] = 'IO'
            i = __unpack_byte_fields(data, packet, 4, 'N DIGIN', 'DIGIN', 'FRONTBUMPS', 'REARBUMPS', 'IRS', 'N_DIGOUT',
                                     'DIGOUT', 'N_AN')
            analogs = []
            for analog in range(data['N_AN']):
                analogs.append(__b_2_i(packet, i))
                i += 2
            data.update({'ANALOGS': analogs})
    except IndexError:
        raise InvalidPacket('ARCOS SIP fields invalid')
    return data


class ARCOSClient:
    """ An ARCOS Client communicating over a serial port with an ARCOS server.

    Args:
        timeout (float): The time to wait while receiving data before a timeout occurs, in seconds.
        write_timeout (float): The time to wait while sending data before a timeout occurs, in seconds.
        allowed_timeouts (int): The number of timeouts to tolerate before the update coroutine closes the port.

    Attributes:
        standard (dict): The last standard Server Information Packet (SIP) received, or `None`, if one hasn't been
            received yet.
        config (dict): The last CONFIGpac SIP received, or `None`, if one hasn't been received yet.
        encoder (dict): The last ENCODERpac SIP received, or `None`, if one hasn't been received yet.
        io (dict): The last IOpac SIP received, or `None`, if one hasn't been received yet.
        standard_event (:class:`threading.Event`): Set whenever a standard SIP is received.
        config_event (:class:`threading.Event`): Set whenever a CONFIGpac SIP is received.
        encoder_event (:class:`threading.Event`): Set whenever an ENCODERpac SIP is received.
        io_event (:class:`threading.Event`): Set whenever an IOpac is received.
        sonars (list): A list of the latest sonar array values, updated whenever a standard SIP is received.
    """
    def __init__(self, timeout=1.0, write_timeout=1.0, allowed_timeouts=2):
        self.timeout = timeout  # Timeout values, in seconds
        self.write_timeout = write_timeout
        # If more than this number of consecutive timeouts occur, close the port, as it is unlikely it is still usable
        self.allowed_timeouts = allowed_timeouts
        self.ser = None  # The serial port, initially nonexistent
        self.serial_lock = Lock()  # A lock is needed so that packets sent by separate threads do not interfere
        self.pulse_running = False
        self.update_running = False
        # Store the last packets received of each type, and make them trigger events
        self.standard = None
        self.standard_event = Event()
        self.config = None
        self.config_event = Event()
        self.encoder = None
        self.encoder_event = Event()
        self.io = None
        self.io_event = Event()
        self.sonars = [5000]*32  # Initialize all possible sonars to 5000 mm

    def send_packet(self, *data):
        """ Send arbitrary data to the ARCOS server.

        Adds the packet header and checksum. Thread-safe.

        Args:
            *data: A tuple or iterable of bytes, whose values are assumed to be between 0 and 255, inclusive.

        Raises:
            `Timeout`: If the write timeout of the serial port was exceeded.
        """
        packet = [0xfa, 0xfb, len(data) + 2] + list(data)  # 0xfa, 0xfb are the packet header
        checksum = packet_checksum(packet)  # Calculate the checksum and append it
        packet.extend([checksum >> 8, checksum & 0xff])  # Big-endian two byte integer
        with self.serial_lock:
            try:
                self.ser.write(bytearray(packet))
            except SerialTimeoutException as e:  # Recast serial timeout as an ARCOS timeout
                raise Timeout(str(e)) from e

    def receive_packet(self):
        """ Read an entire ARCOS Packet from an open port, including header and checksum bytes.

        Returns:
            list: The entire packet as a list of bytes, including header and checksum bytes.

        Raises:
            `Timeout`: If at any point a timeout occurs and fewer bytes than expected are read.
            `InvalidPacket`: If the packet header, checksum, or packet length are invalid.
        """
        def read():
            try:
                b = ord(self.ser.read())
            except TypeError:  # A timeout has occurred, as 0 bytes were read
                raise Timeout
            else:
                return b
        with self.serial_lock:
            # Grab the packet header and ensure it is valid
            h1 = read()
            h2 = read()
            if h1 != 0xfa or h2 != 0xfb:
                raise InvalidPacket('ARCOS header invalid')
            data = [0xfa, 0xfb]
            length = read()
            if length > 249:
              raise InvalidPacket('Invalid packet length')
            data.append(length)
            for i in range(length):
                data.append(read())

        received_crc = (data[-1] & 0xFF) | (data[-2] << 8)
        crc = packet_checksum(data)
        if crc != received_crc:  # Checksum match failure
            raise InvalidPacket('Received checksum ' + str(received_crc) + ', expected checksum ' + str(crc))
        return data

    def send_command(self, code, data=None):
        """ Send a command and data to the ARCOS server.

        Args:
            code: The command code. Must be in :data:`soar.robot.arcos.command_types`.
            data (optional): The associated command argument, assumed to be of the correct type.
                For commands that take a string argument, a `bytes` object may also be used.

        Raises:
            `Timeout`: If the write timeout of the port was exceeded.
        """
        if command_types[code] is None:  # No argument
            self.send_packet(code)
        elif command_types[code] == int:
            if data >= 0:
                arg_type = 0x3b  # Positive integer
            else:
                arg_type = 0x1b  # Negative integer
            data = abs(data)
            b = [data & 0xff, (data >> 8) & 0xff]
            self.send_packet(code, arg_type, *b)
        else:  # command_types[code] == str
            b = list(bytes(data))
            b.append(0)
            arg_type = 0x2b
            self.send_packet(code, arg_type, *b)
            
    def connect(self, forced_ports=None):
        """ Attempt to connect and sync with an ARCOS server over a serial port.

        Returns if successful.

        Args:
            forced_ports (list, optional): If provided, a list of serial ports to try connecting to. Otherwise, the
                client will try all available ports.

        Raises:
            `ARCOSError`: If unable to connect to any available ports.
        """
        if forced_ports is not None:
            ports = forced_ports
        else:
            ports = [port_info.device for port_info in comports()]  # Try every available port until we find a robot
        for port in ports:
            def connect_with_baudrate(baudrate):
                return Serial(port=port, baudrate=baudrate, timeout=self.timeout, writeTimeout=self.write_timeout)
            for baudrate in [115200, 57600, 38400, 19200, 9600]:  # Connect with the highest baudrate possible
                # Attempt to open the port
                try:
                    # Kill the microcontroller servers in case they are already running
                    self.ser = connect_with_baudrate(baudrate)
                    self.send_packet(CLOSE)
                    self.ser.close()

                    # Connect for real, and flush the input and output buffers
                    self.ser = connect_with_baudrate(baudrate)
                    self.ser.reset_input_buffer()
                    self.ser.reset_output_buffer()
                except SerialException as e:  # Any error opening the port (permissions, etc) will cause us to move on
                    printerr('SerialException:', str(e))
                    break
                # Try to sync; if there is a timeout, the port is probably not connected to a robot, so move on
                try:
                    self.sync()
                except Timeout:
                    continue
                else:
                    self.start()
                    return
        # If we have tried every available port without success, raise an exception
        raise ARCOSError('Unable to sync with an ARCOS server--is the robot connected and its port accessible?')

    def disconnect(self):
        """ Stop the ARCOS server and close the connection if running. """
        self.pulse_running = False  # Kill the pulse timer
        self.update_running = False
        if self.ser:  # Only attempt this if the serial port exists
            try:
                self.send_packet(STOP)  # Stop the robot
                self.send_packet(CLOSE)
                self.ser.close()
            except SerialException:  # Ignore errors that occur and assume the port is closed
                pass

    def sync(self, tries=4):
        """ Try to sync with an ARCOS server connected over an open serial port.

        Returns the raw robot identifying information packet sent after `SYNC2` if successful.

        Args:
            tries (int, optional): The number of failures to tolerate before timing out.

        Raises:
            `Timeout`: If the number of tries is exhausted and syncing was not completed.
        """
        for sync in [SYNC0, SYNC1, SYNC2]:
            while True:
                try:
                    self.send_packet(sync)
                    echo = self.receive_packet()
                except Timeout:  # Timeouts just decrement the tries count
                    tries -= 1
                except InvalidPacket:  # Try flushing the input/output buffers
                    self.ser.reset_input_buffer()
                    self.ser.reset_output_buffer()
                    tries -= 1
                else:
                    if sync == echo[3]:
                        break
                    else:
                        tries -= 1
                if tries < 0:
                    raise Timeout('An error occurred while syncing')
        return echo

    def pulse(self):
        """ Continually send the PULSE command so that the robot knows the client is alive. """
        self.pulse_running = True
        while self.pulse_running:
            try:
                self.send_packet(PULSE)
            except Timeout:  # Ignore pulse timeouts, the update coroutine will handle closing the port
                pass
            finally:
                sleep(1.0)  # Default Watchdog interval is 2 seconds, so PULSE every second just to be safe

    def update(self):
        """ Continually receive and decode packets, storing them as attributes and triggering events. """
        self.update_running = True
        timeouts = 0
        invalids = 0
        while self.update_running:
            try:
                received = self.receive_packet()
                decoded = decode_packet(received)
            except Timeout:  # Count timeouts and close the port if too many occur, killing this routine
                timeouts += 1
                if timeouts > self.allowed_timeouts:
                    self.disconnect()
                    break
            except InvalidPacket:  # As per the Pioneer handbook, ignore invalid SIPs and move on
                invalids += 1
                if invalids > 10:  # If many consecutive packets are invalid, try flushing the input buffer
                    self.ser.reset_input_buffer()
                continue
            else:
                invalids = 0
                if decoded['TYPE'] == 'STANDARD':  # Trigger the standard SIP event, and update the latest sonar reading
                    self.standard = decoded
                    for sonar, dist in self.standard['SONARS'].items():
                        self.sonars[sonar] = dist
                    self.standard_event.set()
                elif decoded['TYPE'] == 'CONFIG':
                    self.config = decoded
                    self.config_event.set()
                elif decoded['TYPE'] == 'ENCODER':
                    self.encoder = decoded
                    self.encoder_event.set()
                elif decoded['TYPE'] == 'IO':
                    self.io = decoded
                    self.io_event.set()

    def start(self):
        """ Open the ARCOS servers, enable the sonars, and start the pulse & update coroutines. """
        self.send_packet(OPEN)
        pulse = Thread(target=self.pulse, daemon=True)
        pulse.start()
        update = Thread(target=self.update, daemon=True)
        update.start()
        self.wait_or_timeout(self.standard_event, 5.0, 'Failed to receive SIPs from the robot')
        for i in range(5):  # Try multiple times to enable the sonars
            if self.standard['FLAGS'] & 0x2 != 0x2:
                self.send_command(SONAR, 1)
                sleep(1.0)
            else:
                sleep(1.0)
                return
        if self.standard['FLAGS'] & 0x2 != 0x2:  # If they still aren't enabled, raise an exception
            raise ARCOSError('Unable to enable the robot sonars.')

    @staticmethod
    def wait_or_timeout(event, timeout=1.0, timeout_msg=''):
        """ Wait for an event to occur, with an optional timeout and message.

        Args:
            event (Event): The event to wait for. Expected to be one of the attribute events of this class.
            timeout (float, optional): How long to wait for the event before timing out.
            timeout_msg (str): The message to pass if a timeout occurs.

        Raises:
            `Timeout`: If the event has not occurred by the specified time.
        """
        event_occurred = event.wait(timeout)
        event.clear()
        if not event_occurred:
            raise Timeout(timeout_msg)

